@@ -1,7 +1,7 @@
 /*
  * charconv.c - character code conversion library
  *
- *   Copyright (c) 2000-2003 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2007  Shiro Kawai  <shiro@acm.org>
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,11 +30,12 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: charconv.c,v 1.54 2006/07/19 05:25:10 shirok Exp $
+ *  $Id: charconv.c,v 1.57 2007/03/02 07:39:04 shirok Exp $
  */
 
 #include <string.h>
 #include <errno.h>
+#define GAUCHE_API_0_9
 #include <gauche.h>
 #include <gauche/extend.h>
 #include "charconv.h"
@@ -137,7 +138,7 @@ static ScmObj conv_name(int dir, ScmPort *remote, const char *from, const char *
     Scm_Printf(SCM_PORT(out), "[conv(%s->%s) %s %S]",
                from, to, (dir == SCM_PORT_INPUT? "from" : "to"),
                Scm_PortName(remote));
-    return Scm_GetOutputStringUnsafe(SCM_PORT(out));
+    return Scm_GetOutputStringUnsafe(SCM_PORT(out), 0);
 }
 
 /*------------------------------------------------------------
@@ -214,7 +215,7 @@ static int conv_input_filler(ScmPort *port, int mincnt)
         /* it's likely that the input contains invalid sequence. */
         int cnt = inroom >= 6 ? 6 : inroom;
         ScmObj s = Scm_MakeString(info->buf+insize-inroom, cnt, cnt,
-                                  SCM_MAKSTR_COPYING|SCM_MAKSTR_INCOMPLETE);
+                                  SCM_STRING_COPYING|SCM_STRING_INCOMPLETE);
         Scm_Error("invalid character sequence in the input stream: %S ...", s);
     }
 
