@@ -1,7 +1,7 @@
 /*
  * net.c - network interface
  *
- *   Copyright (c) 2000-2004 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2007  Shiro Kawai  <shiro@acm.org>
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,13 +30,14 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: net.c,v 1.36 2004/08/02 12:20:54 shirok Exp $
+ *  $Id: net.c,v 1.53 2007/03/15 22:43:14 shirok Exp $
  */
 
 #include "gauche/net.h"
 #include <unistd.h>
 #include <fcntl.h>
 #include <gauche/extend.h>
+#include <gauche/uvector.h>
 
 /*==================================================================
  * Socket
@@ -50,11 +51,14 @@ static void socket_finalize(ScmObj obj, void *data)
 {
     ScmSocket *sock = (ScmSocket*)obj;
     /* NB: at this point, sock->inPort and sock->outPort may already
-       be GC-ed and finalized, so we don't flush them here. */
+       be GC-ed and finalized, so we don't flush them here.
+       Clearing them would help them to be collected earlier, though. */
     if (!(SOCKET_CLOSED(sock->fd))) {
         closeSocket(sock->fd);
         sock->fd = INVALID_SOCKET;
         sock->status = SCM_SOCKET_STATUS_CLOSED;
+        sock->inPort = NULL;
+        sock->outPort = NULL;
     }
 }
 
@@ -130,14 +134,28 @@ ScmObj Scm_SocketClose(ScmSocket *s)
     }
     /* We don't shutdown the connection; forked process may have
        reference to the same socket. */
-    if (s->inPort)  Scm_ClosePort(s->inPort);  /* ignore errors */
-    if (s->outPort) Scm_ClosePort(s->outPort); /* ignore errors */
+    /* Clearing inPort/outPort helps them to be collected earlier. */
+    if (s->inPort)  {
+        Scm_ClosePort(s->inPort);  /* ignore errors */
+        s->inPort = NULL;
+    }
+    if (s->outPort) {
+        Scm_ClosePort(s->outPort); /* ignore errors */
+        s->outPort = NULL;
+    }
     closeSocket(s->fd);
     s->fd = INVALID_SOCKET;
     s->status = SCM_SOCKET_STATUS_CLOSED;
     return SCM_TRUE;
 }
 
+static void sockport_err(ScmSocket *sock, const char *io)
+{
+    Scm_Error("attempt to obtain an %s port from unconnected or closed socket: %S",
+              io, sock);
+}
+
+
 ScmObj Scm_SocketInputPort(ScmSocket *sock, int buffering)
 {
     if (sock->inPort == NULL) {
@@ -145,14 +163,15 @@ ScmObj Scm_SocketInputPort(ScmSocket *sock, int buffering)
 	int infd;
         if (sock->type != SOCK_DGRAM &&
             sock->status < SCM_SOCKET_STATUS_CONNECTED) {
-            Scm_Error("attempt to obtain an input port from unconnected socket: %S",
-                      SCM_OBJ(sock));
+            sockport_err(sock, "input");
         }
 #ifndef __MINGW32__
 	infd = sock->fd;
 #else  /*__MINGW32__*/
 	infd = _open_osfhandle(sock->fd, O_RDONLY);
 #endif /*__MINGW32__*/
+        if (infd == INVALID_SOCKET) sockport_err(sock, "input");
+        
         /* NB: I keep the socket itself in the port name, in order to avoid
            the socket from GCed prematurely if application doesn't keep
            pointer to the socket. */
@@ -171,14 +190,14 @@ ScmObj Scm_SocketOutputPort(ScmSocket *sock, int buffering)
 	int outfd;
         if (sock->type != SOCK_DGRAM &&
             sock->status < SCM_SOCKET_STATUS_CONNECTED) {
-            Scm_Error("attempt to obtain an output port from an unconnected socket: %S",
-                      SCM_OBJ(sock));
+            sockport_err(sock, "output");
         }
 #ifndef __MINGW32__
 	outfd = sock->fd;
 #else  /*__MINGW32__*/
 	outfd = _open_osfhandle(sock->fd, 0);
 #endif /*__MINGW32__*/
+        if (outfd == INVALID_SOCKET) sockport_err(sock, "output");
 
         /* NB: I keep the socket itself in the port name, in order to avoid
            the socket from GCed prematurely if application doesn't keep
@@ -194,14 +213,18 @@ ScmObj Scm_SocketOutputPort(ScmSocket *sock, int buffering)
  * Low-level library
  */
 
+#define CLOSE_CHECK(fd, op, s)                                          \
+    do {                                                                \
+        if (SOCKET_CLOSED(fd)) {                                        \
+            Scm_Error("attempt to %s a closed socket: %S", op, s);      \
+        }                                                               \
+    } while (0)
+        
 ScmObj Scm_SocketBind(ScmSocket *sock, ScmSockAddr *addr)
 {
     ScmSockAddr *naddr;
     int r;
-    
-    if (SOCKET_CLOSED(sock->fd)) {
-        Scm_Error("attempt to bind a closed socket: %S", sock);
-    }
+    CLOSE_CHECK(sock->fd, "bind", sock);
     SCM_SYSCALL(r, bind(sock->fd, &addr->addr, addr->addrlen));
     if (r < 0) {
         Scm_SysError("bind failed to %S", addr);
@@ -223,9 +246,7 @@ ScmObj Scm_SocketBind(ScmSocket *sock, ScmSockAddr *addr)
 ScmObj Scm_SocketListen(ScmSocket *sock, int backlog)
 {
     int r;
-    if (SOCKET_CLOSED(sock->fd)) {
-        Scm_Error("attempt to listen a closed socket: %S", sock);
-    }
+    CLOSE_CHECK(sock->fd, "listen to", sock);
     SCM_SYSCALL(r, listen(sock->fd, backlog));
     if (r < 0) {
         Scm_SysError("listen(2) failed");
@@ -236,15 +257,14 @@ ScmObj Scm_SocketListen(ScmSocket *sock, int backlog)
 
 ScmObj Scm_SocketAccept(ScmSocket *sock)
 {
-    const char addrbuf[SCM_SOCKADDR_MAXLEN];
-    int newfd, addrlen = SCM_SOCKADDR_MAXLEN;
+    int newfd;
+    struct sockaddr_storage addrbuf;
+    socklen_t addrlen = sizeof(addrbuf);
     ScmSocket *newsock;
     ScmClass *addrClass = Scm_ClassOf(SCM_OBJ(sock->address));
-    
-    if (SOCKET_CLOSED(sock->fd)) {
-        Scm_Error("attempt to accept a closed socket: %S", sock);
-    }
-    SCM_SYSCALL(newfd, accept(sock->fd, (struct sockaddr *)addrbuf, &addrlen));
+
+    CLOSE_CHECK(sock->fd, "accept from", sock);
+    SCM_SYSCALL(newfd, accept(sock->fd, (struct sockaddr*)&addrbuf, &addrlen));
     if (SOCKET_INVALID(newfd)) {
         if (errno == EAGAIN) {
             return SCM_FALSE;
@@ -255,7 +275,7 @@ ScmObj Scm_SocketAccept(ScmSocket *sock)
     newsock = make_socket(newfd, sock->type);
     newsock->address =
         SCM_SOCKADDR(Scm_MakeSockAddr(addrClass,
-                                      (struct sockaddr *)addrbuf,
+                                      (struct sockaddr*)&addrbuf,
                                       addrlen));
     newsock->status = SCM_SOCKET_STATUS_CONNECTED;
     return SCM_OBJ(newsock);
@@ -264,9 +284,7 @@ ScmObj Scm_SocketAccept(ScmSocket *sock)
 ScmObj Scm_SocketConnect(ScmSocket *sock, ScmSockAddr *addr)
 {
     int r;
-    if (SOCKET_CLOSED(sock->fd)) {
-        Scm_Error("attempt to connect a closed socket: %S", sock);
-    }
+    CLOSE_CHECK(sock->fd, "connect to", sock);
     SCM_SYSCALL(r, connect(sock->fd, &addr->addr, addr->addrlen));
     if (r < 0) {
         Scm_SysError("connect failed to %S", addr);
@@ -278,59 +296,67 @@ ScmObj Scm_SocketConnect(ScmSocket *sock, ScmSockAddr *addr)
 
 ScmObj Scm_SocketGetSockName(ScmSocket *sock)
 {
-    const char addrbuf[SCM_SOCKADDR_MAXLEN];
-    int r, addrlen = SCM_SOCKADDR_MAXLEN;
+    int r;
+    struct sockaddr_storage addrbuf;
+    socklen_t addrlen = sizeof(addrbuf);
 
-    if (SOCKET_CLOSED(sock->fd)) {
-        Scm_Error("attempt to get the name of a closed socket: %S", sock);
-    }
-    SCM_SYSCALL(r, getsockname(sock->fd, (struct sockaddr *)addrbuf, &addrlen));
+    CLOSE_CHECK(sock->fd, "get the name of", sock);
+    SCM_SYSCALL(r, getsockname(sock->fd, (struct sockaddr*)&addrbuf, &addrlen));
     if (r < 0) {
         Scm_SysError("getsockname(2) failed");
     }
-    return SCM_OBJ(Scm_MakeSockAddr(NULL, (struct sockaddr *)addrbuf, addrlen));
+    return SCM_OBJ(Scm_MakeSockAddr(NULL, (struct sockaddr*)&addrbuf, addrlen));
 }
 
 ScmObj Scm_SocketGetPeerName(ScmSocket *sock)
 {
-    const char addrbuf[SCM_SOCKADDR_MAXLEN];
-    int r, addrlen = SCM_SOCKADDR_MAXLEN;
+    int r;
+    struct sockaddr_storage addrbuf;
+    socklen_t addrlen = sizeof(addrbuf);
 
-    if (SOCKET_CLOSED(sock->fd)) {
-        Scm_Error("attempt to get the name of a closed socket: %S", sock);
-    }
-    SCM_SYSCALL(r, getpeername(sock->fd, (struct sockaddr *)addrbuf, &addrlen));
+    CLOSE_CHECK(sock->fd, "get the name of", sock);
+    SCM_SYSCALL(r, getpeername(sock->fd, (struct sockaddr*)&addrbuf, &addrlen));
     if (r < 0) {
         Scm_SysError("getpeername(2) failed");
     }
-    return SCM_OBJ(Scm_MakeSockAddr(NULL, (struct sockaddr *)addrbuf, addrlen));
+    return SCM_OBJ(Scm_MakeSockAddr(NULL, (struct sockaddr*)&addrbuf, addrlen));
 }
 
-ScmObj Scm_SocketSend(ScmSocket *sock, ScmString *msg, int flags)
+static const char *get_message_body(ScmObj msg, u_int *size)
 {
-    int r;
-    if (SOCKET_CLOSED(sock->fd)) {
-        Scm_Error("attempt to send to a closed socket: %S", sock);
+    if (SCM_UVECTORP(msg)) {
+        *size = Scm_UVectorSizeInBytes(SCM_UVECTOR(msg));
+        return (const char*)SCM_UVECTOR_ELEMENTS(msg);
+    } else if (SCM_STRINGP(msg)) {
+        return Scm_GetStringContent(SCM_STRING(msg), size, NULL, NULL);
+    } else {
+        Scm_TypeError("socket message", "uniform vector or string", msg);
+        *size = 0;              /* dummy */
+        return NULL;
     }
-    SCM_SYSCALL(r, send(sock->fd,
-                        SCM_STRING_START(msg), SCM_STRING_SIZE(msg),
-                        flags));
+}
+
+ScmObj Scm_SocketSend(ScmSocket *sock, ScmObj msg, int flags)
+{
+    int r; u_int size;
+    const char *cmsg;
+    CLOSE_CHECK(sock->fd, "send to", sock);
+    cmsg = get_message_body(msg, &size);
+    SCM_SYSCALL(r, send(sock->fd, cmsg, size, flags));
     if (r < 0) {
         Scm_SysError("send(2) failed");
     }
     return SCM_MAKE_INT(r);
 }
 
-ScmObj Scm_SocketSendTo(ScmSocket *sock, ScmString *msg, ScmSockAddr *to,
+ScmObj Scm_SocketSendTo(ScmSocket *sock, ScmObj msg, ScmSockAddr *to,
                         int flags)
 {
-    int r;
-    if (SOCKET_CLOSED(sock->fd)) {
-        Scm_Error("attempt to send to a closed socket: %S", sock);
-    }
-    SCM_SYSCALL(r, sendto(sock->fd,
-                          SCM_STRING_START(msg), SCM_STRING_SIZE(msg),
-                          flags,
+    int r; u_int size;
+    const char *cmsg;
+    CLOSE_CHECK(sock->fd, "send to", sock);
+    cmsg = get_message_body(msg, &size);
+    SCM_SYSCALL(r, sendto(sock->fd, cmsg, size, flags,
                           &SCM_SOCKADDR(to)->addr, SCM_SOCKADDR(to)->addrlen));
     if (r < 0) {
         Scm_SysError("sendto(2) failed");
@@ -342,33 +368,93 @@ ScmObj Scm_SocketRecv(ScmSocket *sock, int bytes, int flags)
 {
     int r;
     char *buf;
-    if (SOCKET_CLOSED(sock->fd)) {
-        Scm_Error("attempt to recv from a closed socket: %S", sock);
-    }
+    CLOSE_CHECK(sock->fd, "recv from", sock);
     buf = SCM_NEW_ATOMIC2(char*, bytes);
     SCM_SYSCALL(r, recv(sock->fd, buf, bytes, flags));
     if (r < 0) {
         Scm_SysError("recv(2) failed");
     }
-    return Scm_MakeString(buf, r, r, SCM_MAKSTR_INCOMPLETE);
+    return Scm_MakeString(buf, r, r, SCM_STRING_INCOMPLETE);
+}
+
+static char *get_message_buffer(ScmUVector *v, u_int *size)
+{
+    if (SCM_UVECTOR_IMMUTABLE_P(v)) {
+        Scm_Error("attempted to use an immutable uniform vector as a buffer");
+    }
+    *size = Scm_UVectorSizeInBytes(v);
+    return (char *)SCM_UVECTOR_ELEMENTS(v);
+}
+
+ScmObj Scm_SocketRecvX(ScmSocket *sock, ScmUVector *buf, int flags)
+{
+    int r;
+    u_int size;
+    char *z;
+    CLOSE_CHECK(sock->fd, "recv from", sock);
+    z = get_message_buffer(buf, &size);
+    SCM_SYSCALL(r, recv(sock->fd, z, size, flags));
+    if (r < 0) {
+        Scm_SysError("recv(2) failed");
+    }
+    return Scm_MakeInteger(r);
 }
 
 ScmObj Scm_SocketRecvFrom(ScmSocket *sock, int bytes, int flags)
 {
     int r;
     char *buf;
-    struct sockaddr from;
-    int fromlen = sizeof(from);
-    if (SOCKET_CLOSED(sock->fd)) {
-        Scm_Error("attempt to recv from a closed socket: %S", sock);
-    }
+    struct sockaddr_storage from;
+    socklen_t fromlen = sizeof(from);
+    CLOSE_CHECK(sock->fd, "recv from", sock);
     buf = SCM_NEW_ATOMIC2(char*, bytes);
-    SCM_SYSCALL(r, recvfrom(sock->fd, buf, bytes, flags, &from, &fromlen));
+    SCM_SYSCALL(r, recvfrom(sock->fd, buf, bytes, flags,
+                            (struct sockaddr*)&from, &fromlen));
     if (r < 0) {
         Scm_SysError("recvfrom(2) failed");
     }
-    return Scm_Values2(Scm_MakeString(buf, r, r, SCM_MAKSTR_INCOMPLETE),
-                       Scm_MakeSockAddr(NULL, &from, fromlen));
+    return Scm_Values2(Scm_MakeString(buf, r, r, SCM_STRING_INCOMPLETE),
+                       Scm_MakeSockAddr(NULL, (struct sockaddr*)&from, fromlen));
+}
+
+/* ADDRS is a list of socket addresses; if 'from' address type matches
+   one of them, it is used to store the information so that we can avoid
+   allocation.  If no addresses match the incoming type, and ADDRS is
+   a complete list, the information of 'from' is discarded.  If no addresses
+   match the incoming type, and the last cdr of ADDRS is #t (this case
+   includes ADDRS == #t), a new sockaddr is allocated and returned. */
+ScmObj Scm_SocketRecvFromX(ScmSocket *sock, ScmUVector *buf,
+                           ScmObj addrs, int flags)
+{
+    int r;
+    u_int size;
+    char *z;
+    struct sockaddr_storage from;
+    socklen_t fromlen = sizeof(from);
+    ScmObj cp, addr = SCM_FALSE;
+    
+    CLOSE_CHECK(sock->fd, "recv from", sock);
+    z = get_message_buffer(buf, &size);
+    SCM_SYSCALL(r, recvfrom(sock->fd, z, size, flags,
+                            (struct sockaddr*)&from, &fromlen));
+    if (r < 0) {
+        Scm_SysError("recvfrom(2) failed");
+    }
+    SCM_FOR_EACH(cp, addrs) {
+        ScmObj a = SCM_CAR(cp);
+        if (Scm_SockAddrP(a)) {
+            if (SCM_SOCKADDR_FAMILY(a) == from.ss_family) {
+                memcpy(&SCM_SOCKADDR(a)->addr, &from, SCM_SOCKADDR(a)->addrlen);
+                addr = a;
+                break;
+            }
+        }
+    }
+    if (SCM_FALSEP(addr) && SCM_EQ(cp, SCM_TRUE)) {
+        /* Allocate sockaddr */
+        addr = Scm_MakeSockAddr(NULL, (struct sockaddr*)&from, fromlen);
+    }
+    return Scm_Values2(Scm_MakeInteger(r), addr);
 }
 
 /* Low-level setsockopt() and getsockopt() interface. */
@@ -380,17 +466,23 @@ ScmObj Scm_SocketRecvFrom(ScmSocket *sock, int bytes, int flags)
 ScmObj Scm_SocketSetOpt(ScmSocket *s, int level, int option, ScmObj value)
 {
     int r = 0;
-    if (SOCKET_CLOSED(s->fd)) {
-        Scm_Error("attempt to set a socket option of a closed socket: %S", s);
-    }
+    CLOSE_CHECK(s->fd, "set a socket option of", s);
     if (SCM_STRINGP(value)) {
-        SCM_SYSCALL(r, setsockopt(s->fd, level, option, SCM_STRING_START(value),
-                                    SCM_STRING_SIZE(value)));
+        u_int size;
+        const char *cvalue = Scm_GetStringContent(SCM_STRING(value), &size,
+                                                  NULL, NULL);
+        SCM_SYSCALL(r, setsockopt(s->fd, level, option, cvalue, size));
+    } else if (SCM_UVECTORP(value)) {
+        u_int size = Scm_UVectorSizeInBytes(SCM_UVECTOR(value));
+        const char *cvalue = (const char*)SCM_UVECTOR_ELEMENTS(value);
+        SCM_SYSCALL(r, setsockopt(s->fd, level, option, cvalue, size));
     } else if (SCM_INTP(value) || SCM_BIGNUMP(value)) {
         int v = Scm_GetInteger(value);
-        SCM_SYSCALL(r, setsockopt(s->fd, level, option, &v, sizeof(int)));
+        SCM_SYSCALL(r, setsockopt(s->fd, level, option, (void*)&v, sizeof(int)));
     } else {
-        Scm_Error("socket option must be a string or an integer: %S", value);
+        Scm_TypeError("socket option value",
+                      "an integer, a uvector or a string",
+                      value);
     }
     if (r < 0) Scm_SysError("setsockopt failed");
     return SCM_TRUE;
@@ -399,18 +491,17 @@ ScmObj Scm_SocketSetOpt(ScmSocket *s, int level, int option, ScmObj value)
 ScmObj Scm_SocketGetOpt(ScmSocket *s, int level, int option, int rsize)
 {
     int r = 0;
-    if (SOCKET_CLOSED(s->fd)) {
-        Scm_Error("attempt to get a socket option of a closed socket: %S", s);
-    }
+    socklen_t rrsize = rsize;
+    CLOSE_CHECK(s->fd, "get a socket option of", s);
     if (rsize > 0) {
-        char *buf = SCM_NEW_ATOMIC2(char *, rsize);
-        SCM_SYSCALL(r, getsockopt(s->fd, level, option, buf, &rsize));
+        char *buf = SCM_NEW_ATOMIC2(char *, rrsize);
+        SCM_SYSCALL(r, getsockopt(s->fd, level, option, buf, &rrsize));
         if (r < 0) Scm_SysError("getsockopt failed");
-        return Scm_MakeString(buf, rsize, rsize, SCM_MAKSTR_INCOMPLETE);
+        return Scm_MakeString(buf, rrsize, rrsize, SCM_STRING_INCOMPLETE);
     } else {
         int val;
-        rsize = sizeof(int);
-        SCM_SYSCALL(r, getsockopt(s->fd, level, option, &val, &rsize));
+        rrsize = sizeof(int);
+        SCM_SYSCALL(r, getsockopt(s->fd, level, option, (void*)&val, &rrsize));
         if (r < 0) Scm_SysError("getsockopt failed");
         return Scm_MakeInteger(val);
     }
@@ -438,6 +529,28 @@ int inet_aton(const char *cp, struct in_addr *inp)
 /* winsock requires some obscure initialization */
 static WSADATA wsaData;
 
+static void init_winwock(void)
+{
+    int opt;
+    int r = WSAStartup(MAKEWORD(2,2), &wsaData);
+    if (r != 0) {
+        SetLastError(r);
+        Scm_SysError("WSAStartup failed");
+    }
+    /* windows voodoo to make _open_osfhandle magic work */
+    opt = SO_SYNCHRONOUS_NONALERT;
+    r = setsockopt(INVALID_SOCKET, SOL_SOCKET,
+                   SO_OPENTYPE, (char*)&opt, sizeof(opt));
+    if (r == SOCKET_ERROR) {
+        Scm_SysError("winsock initialization failed");
+    }
+}
+
+static void fini_winsock(void *data)
+{
+    (void)WSACleanup();
+}
+
 #endif /*__MINGW32__*/
                           
 /*==================================================================
@@ -447,93 +560,27 @@ static WSADATA wsaData;
 extern void Scm_Init_NetAddr(ScmModule *mod);
 extern void Scm_Init_NetDB(ScmModule *mod);
 extern void Scm_Init_netlib(ScmModule *mod);
+extern void Scm_Init_netaux(void);
+
+
+
 
 void Scm_Init_libnet(void)
 {
     ScmModule *mod;
 
     SCM_INIT_EXTENSION(net);
-    mod = SCM_MODULE(SCM_FIND_MODULE("gauche.net", TRUE));
+    mod = SCM_FIND_MODULE("gauche.net", SCM_FIND_MODULE_CREATE);
 #ifdef __MINGW32__
-    /* NB: I'm supposed to call WSACleanup when application shuts down,
-       or resource leak would happen, according to the Windows document.
-       It's just ridiculous---why can't OS itself clean up the dead process?
-       Anyway, to behave politically correctly, I need a generic callback
-       mechanism for Gauche core to call cleanup routines in Scm_Exit.
-       For now, let us behave like a wild kid.
-    */
-    {
-	int opt;
-	int r = WSAStartup(MAKEWORD(2,2), &wsaData);
-	if (r != 0) {
-	    SetLastError(r);
-	    Scm_SysError("WSAStartup failed");
-	}
-	/* windows voodoo to make _open_osfhandle magic work */
-	opt = SO_SYNCHRONOUS_NONALERT;
-	r = setsockopt(INVALID_SOCKET, SOL_SOCKET,
-		       SO_OPENTYPE, (char*)&opt, sizeof(opt));
-	if (r == SOCKET_ERROR) {
-	    Scm_SysError("winsock initialization failed");
-	}
-    }
+    init_winsock();
+    Scm_AddCleanupHandler(fini_winsock, NULL);
 #endif /*__MINGW32__*/
-    Scm_InitBuiltinClass(&Scm_SocketClass, "<socket>", NULL,
-                         sizeof(ScmSocket), mod);
+#ifdef HAVE_IPV6
+    Scm_AddFeature("gauche.net.ipv6", NULL);
+#endif
+    Scm_InitStaticClass(&Scm_SocketClass, "<socket>", mod, NULL, 0);
     Scm_Init_NetAddr(mod);
     Scm_Init_NetDB(mod);
     Scm_Init_netlib(mod);
-
-    /* Constants for socket option operation.
-       I define them here, instead of netlib.stub,  so that I can check
-       if the symbol is defined */
-#define DEFSYM(sym, val) \
-    SCM_DEFINE(mod, sym, Scm_MakeInteger(val))
-
-    DEFSYM("SOL_SOCKET", SOL_SOCKET);
-#ifdef SO_KEEPALIVE
-    DEFSYM("SO_KEEPALIVE", SO_KEEPALIVE);
-#endif
-#ifdef SO_OOBINLINE
-    DEFSYM("SO_OOBINLINE", SO_OOBINLINE);
-#endif
-#ifdef SO_REUSEADDR
-    DEFSYM("SO_REUSEADDR", SO_REUSEADDR);
-#endif
-#ifdef SO_TYPE
-    DEFSYM("SO_TYPE",      SO_TYPE);
-#endif
-#ifdef SO_BROADCAST
-    DEFSYM("SO_BROADCAST", SO_BROADCAST);
-#endif
-#ifdef SO_SNDBUF
-    DEFSYM("SO_SNDBUF",    SO_SNDBUF);
-#endif
-#ifdef SO_RCVBUF
-    DEFSYM("SO_RCVBUF",    SO_RCVBUF);
-#endif
-#ifdef SO_PRIORITY
-    DEFSYM("SO_PRIORITY",  SO_PRIORITY);
-#endif
-    DEFSYM("SO_ERROR",     SO_ERROR);
-
-#ifdef SOL_TCP
-    DEFSYM("SOL_TCP", SOL_TCP);
-#ifdef TCP_NODELAY
-    DEFSYM("TCP_NODELAY",  TCP_NODELAY);
-#endif
-#ifdef TCP_MAXSEG
-    DEFSYM("TCP_MAXSEG",   TCP_MAXSEG);
-#endif
-#ifdef TCP_CORK
-    DEFSYM("TCP_CORK",     TCP_CORK);
-#endif
-#endif /* SOL_TCP */
-
-#ifdef SOL_IP
-    DEFSYM("SOL_IP", SOL_IP);
-#ifdef IP_OPTIONS
-    DEFSYM("IP_OPTIONS",   IP_OPTIONS);
-#endif
-#endif /* SOL_IP */
+    Scm_Init_netaux();
 }
