@@ -1,7 +1,7 @@
 /*
  * netdb.c - obtain information about network
  *
- *   Copyright (c) 2000-2004 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2007  Shiro Kawai  <shiro@acm.org>
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: netdb.c,v 1.10 2004/08/02 12:20:54 shirok Exp $
+ *  $Id: netdb.c,v 1.16 2007/03/02 07:39:06 shirok Exp $
  */
 
 #include "gauche/net.h"
@@ -44,7 +44,7 @@ static struct netdb_data_rec {
     ScmInternalMutex hostent_mutex;
     ScmInternalMutex protoent_mutex;
     ScmInternalMutex servent_mutex;
-} netdb_data = { 0 };
+} netdb_data = { 1 };
 
 #define WITH_GLOBAL_LOCK(mutex, body)           \
     SCM_UNWIND_PROTECT {                        \
@@ -70,18 +70,15 @@ static ScmSysHostent *make_hostent(struct hostent *he)
     
     SCM_SET_CLASS(entry, SCM_CLASS_SYS_HOSTENT);
     entry->name = SCM_MAKE_STR_COPYING(he->h_name);
-    for (p = he->h_aliases; *p; p++) {
-        SCM_APPEND1(h, t, SCM_MAKE_STR_COPYING(*p));
-    }
-    entry->aliases = h;
-    h = t = SCM_NIL;
+    entry->aliases = Scm_CStringArrayToList((const char**)he->h_aliases, -1,
+                                            SCM_STRING_COPYING);
     if (he->h_addrtype == AF_INET) {
         for (p = he->h_addr_list; *p; p++) {
             char buf[50];
             struct in_addr *addr = (struct in_addr*)*p;
             unsigned long addrval = ntohl(addr->s_addr);
             /* avoid using inet_ntoa, for it is not reentrant */
-            snprintf(buf, 50, "%d.%d.%d.%d",
+            snprintf(buf, 50, "%ld.%ld.%ld.%ld",
                      ((addrval >> 24)& 0xff), ((addrval >> 16) & 0xff),
                      ((addrval >> 8) & 0xff), (addrval & 0xff));
             SCM_APPEND1(h, t, SCM_MAKE_STR_COPYING(buf));
@@ -226,15 +223,11 @@ SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_SysProtoentClass, NULL);
 static ScmSysProtoent *make_protoent(struct protoent *pe)
 {
     ScmSysProtoent *entry = SCM_NEW(ScmSysProtoent);
-    ScmObj h = SCM_NIL, t = SCM_NIL;
-    char **p;
     
     SCM_SET_CLASS(entry, SCM_CLASS_SYS_PROTOENT);
     entry->name = SCM_MAKE_STR_COPYING(pe->p_name);
-    for (p = pe->p_aliases; *p; p++) {
-        SCM_APPEND1(h, t, SCM_MAKE_STR_COPYING(*p));
-    }
-    entry->aliases = h;
+    entry->aliases = Scm_CStringArrayToList((const char **)pe->p_aliases, -1,
+                                            SCM_STRING_COPYING);
     entry->proto = Scm_MakeInteger(pe->p_proto);
     return entry;
 }
@@ -350,15 +343,11 @@ SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_SysServentClass, NULL);
 static ScmSysServent *make_servent(struct servent *se)
 {
     ScmSysServent *entry = SCM_NEW(ScmSysServent);
-    ScmObj h = SCM_NIL, t = SCM_NIL;
-    char **p;
     
     SCM_SET_CLASS(entry, SCM_CLASS_SYS_SERVENT);
     entry->name = SCM_MAKE_STR_COPYING(se->s_name);
-    for (p = se->s_aliases; *p; p++) {
-        SCM_APPEND1(h, t, SCM_MAKE_STR_COPYING(*p));
-    }
-    entry->aliases = h;
+    entry->aliases = Scm_CStringArrayToList((const char **)se->s_aliases, -1,
+                                            SCM_STRING_COPYING);
     entry->port = Scm_MakeInteger(ntohs(se->s_port));
     entry->proto = SCM_MAKE_STR_COPYING(se->s_proto);
     return entry;
@@ -545,12 +534,12 @@ ScmObj Scm_GetNameinfo(ScmSockAddr *addr, int flags)
 
 void Scm_Init_NetDB(ScmModule *mod)
 {
-    Scm_InitBuiltinClass(&Scm_SysHostentClass, "<sys-hostent>",
-                         hostent_slots, sizeof(ScmSysHostent), mod);
-    Scm_InitBuiltinClass(&Scm_SysProtoentClass, "<sys-protoent>",
-                         protoent_slots, sizeof(ScmSysProtoent), mod);
-    Scm_InitBuiltinClass(&Scm_SysServentClass, "<sys-servent>",
-                         servent_slots, sizeof(ScmSysServent), mod);
+    Scm_InitStaticClass(&Scm_SysHostentClass, "<sys-hostent>", mod,
+                        hostent_slots, 0);
+    Scm_InitStaticClass(&Scm_SysProtoentClass, "<sys-protoent>", mod,
+                        protoent_slots, 0);
+    Scm_InitStaticClass(&Scm_SysServentClass, "<sys-servent>", mod,
+                        servent_slots, 0);
     SCM_INTERNAL_MUTEX_INIT(netdb_data.hostent_mutex);
     SCM_INTERNAL_MUTEX_INIT(netdb_data.protoent_mutex);
     SCM_INTERNAL_MUTEX_INIT(netdb_data.servent_mutex);
