@@ -1,7 +1,7 @@
 ;;;
 ;;; collection.scm - collection generics
 ;;;  
-;;;   Copyright (c) 2000-2004 Shiro Kawai, All rights reserved.
+;;;   Copyright (c) 2000-2007  Shiro Kawai  <shiro@acm.org>
 ;;;   
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
 ;;;   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 ;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ;;;  
-;;;  $Id: collection.scm,v 1.3 2006/01/07 11:36:16 shirok Exp $
+;;;  $Id: collection.scm,v 1.7 2007/03/02 07:39:12 shirok Exp $
 ;;;
 
 ;; Defines generic operations over collection.   A collection is
@@ -93,15 +93,24 @@
               (begin0 (cons k v)
                       (set!-values (k v) (iter eof-marker))))))))
 
+(define-method call-with-iterator ((coll <tree-map>) proc . args)
+  (let ((eof-marker (cons #f #f))
+        (iter (%tree-map-iter coll)))
+    (receive (k v) (iter eof-marker #f)
+      (proc (cut eq? k eof-marker)
+            (lambda ()
+              (begin0 (cons k v)
+                      (set!-values (k v) (iter eof-marker #f))))))))
+
 ;; n-ary case aux. proc
 (define (call-with-iterators colls proc)
   (let loop ((colls colls)
              (eprocs '())
              (nprocs '()))
     (if (null? colls)
-        (proc (reverse! eprocs) (reverse! nprocs))
-        (with-iterator ((car colls) end? next)
-          (loop (cdr colls) (cons end? eprocs) (cons next nprocs))))))
+      (proc (reverse! eprocs) (reverse! nprocs))
+      (with-iterator ((car colls) end? next)
+        (loop (cdr colls) (cons end? eprocs) (cons next nprocs))))))
 
 ;;-------------------------------------------------
 ;; Call-with-builder - the fundamental constructor
@@ -167,6 +176,16 @@
             (hash-table-put! h (car item) (cdr item)))
           (lambda () h))))
 
+(define-method call-with-builder ((class <tree-map-meta>) proc . args)
+  (let-keywords* args ((key=? =)
+                       (key<? <))
+    (let1 tree (make-tree-map key=? key<?)
+      (proc (lambda (item)
+              (unless (pair? item)
+                (error "pair required to build a tree-map, but got" item))
+              (tree-map-put! tree (car item) (cdr item)))
+            (lambda () tree)))))
+
 ;; utility.  return minimum size of collections if it's easily known, or #f.
 (define (maybe-minimum-size col more)
   (let1 size (and-let* ((siz (lazy-size-of col))
@@ -213,15 +232,15 @@
          )))
     ))
                  
-;; generic way.   This effectively shadows SRFI-1 fold.
+;; generic way.   This shadows builtin fold.
 (define-fold-k fold (knil))
 
-;; for list arguments, SRFI-1 implementation is slightly faster.
+;; for list arguments, builtin fold is faster.
 (define-method fold (proc knil (coll <list>))
-  ((with-module srfi-1 fold) proc knil coll))
+  ((with-module gauche fold) proc knil coll))
 
 (define-method fold (proc knil (coll <list>) (coll2 <list>))
-  ((with-module srfi-1 fold) proc knil coll coll2))
+  ((with-module gauche fold) proc knil coll coll2))
 
 ;; 2- and 3- seed values
 (define-fold-k fold2 (knil1 knil2))
@@ -480,8 +499,8 @@
 ;;  cf. group-sequence in gauche.sequence
 
 (define-method group-collection ((col <collection>) . args)
-  (let-keywords* args ((key-proc  :key identity)
-                       (test-proc :test  eqv?))
+  (let-keywords args ((key-proc  :key identity)
+                      (test-proc :test  eqv?))
     (fold (lambda (b r) (cons (reverse! (cdr b)) r))
           '()
           (fold (lambda (elt buckets)
