@@ -1,7 +1,7 @@
 /*
  * bignum.c - multiple precision exact integer arithmetic
  *
- *   Copyright (c) 2000-2004 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2007  Shiro Kawai  <shiro@acm.org>
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: bignum.c,v 1.56 2004/06/17 23:42:37 shirok Exp $
+ *  $Id: bignum.c,v 1.63 2007/03/02 07:39:12 shirok Exp $
  */
 
 /* Bignum library.  Not optimized well yet---I think bignum performance
@@ -73,6 +73,7 @@ char *alloca ();
 # endif
 #endif
 
+#include <stdlib.h>
 #include <math.h>
 #include <limits.h>
 #define LIBGAUCHE_BODY
@@ -90,8 +91,6 @@ static int bignum_safe_size_for_add(ScmBignum *x, ScmBignum *y);
 static ScmBignum *bignum_add_int(ScmBignum *br, ScmBignum *bx, ScmBignum *by);
 static ScmBignum *bignum_2scmpl(ScmBignum *br);
 
-int Scm_DumpBignum(ScmBignum *b, ScmPort *out);
-
 /*---------------------------------------------------------------------
  * Constructor
  *
@@ -245,56 +244,67 @@ ScmObj Scm_NormalizeBignum(ScmBignum *b)
 }
 
 /* b must be normalized.  */
-long Scm_BignumToSI(ScmBignum *b, int clamphi, int clamplo)
+long Scm_BignumToSI(ScmBignum *b, int clamp, int *oor)
 {
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (b->sign >= 0) {
         if (b->values[0] > LONG_MAX || b->size >= 2) {
-            if (clamphi) return LONG_MAX;
+            if (clamp & SCM_CLAMP_HI) return LONG_MAX;
             else goto err;
         } else {
             return (long)b->values[0];
         }
     } else {
         if (b->values[0] > (u_long)LONG_MAX+1 || b->size >= 2) {
-            if (clamplo) return LONG_MIN;
+            if (clamp & SCM_CLAMP_LO) return LONG_MIN;
             else goto err;
         } else {
             return -(long)b->values[0];
         }
     }
   err:
-    Scm_Error("argument out of range: %S", SCM_OBJ(b));
-    return 0; /*dummy*/
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) {
+        *oor = TRUE;
+    } else {
+        Scm_Error("argument out of range: %S", SCM_OBJ(b));
+    }
+    return 0;
 }
 
 /* b must be normalized. */
-u_long Scm_BignumToUI(ScmBignum *b, int clamphi, int clamplo)
+u_long Scm_BignumToUI(ScmBignum *b, int clamp, int *oor)
 {
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (b->sign >= 0) {
         if (b->size >= 2) {
-            if (clamphi) return SCM_ULONG_MAX;
+            if (clamp & SCM_CLAMP_HI) return SCM_ULONG_MAX;
             else goto err;
         } else {
             return b->values[0];
         }
     } else {
-        if (clamplo) return 0;
+        if (clamp & SCM_CLAMP_LO) return 0;
         else goto err;
     }
   err:
-    Scm_Error("argument out of range: %S", SCM_OBJ(b));
-    return 0; /*dummy*/
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) {
+        *oor = TRUE;
+    } else {
+        Scm_Error("argument out of range: %S", SCM_OBJ(b));
+    }
+    return 0;
 }
 
 #if SIZEOF_LONG == 4
 /* we need special routines for int64 */
-ScmInt64 Scm_BignumToSI64(ScmBignum *b, int clamphi, int clamplo)
+ScmInt64 Scm_BignumToSI64(ScmBignum *b, int clamp, int *oor)
 {
 #if SCM_EMULATE_INT64
     ScmInt64 r = {0, 0};
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (b->sign > 0) {
         if (b->size > 2 || b->values[1] > LONG_MAX) {
-            if (!clamphi) goto err;
+            if (!(clamp & SCM_CLAMP_HI)) goto err;
             SCM_SET_INT64_MAX(r);
         } else {
             r.lo = b->values[0];
@@ -302,7 +312,7 @@ ScmInt64 Scm_BignumToSI64(ScmBignum *b, int clamphi, int clamplo)
         }
     } else if (b->sign < 0) {
         if (b->size > 2 || b->values[1] > (u_long)LONG_MAX + 1) {
-            if (!clamplo) goto err;
+            if (!(clamp&SCM_CLAMP_LO)) goto err;
             SCM_SET_INT64_MIN(r);
         } else {
             b = SCM_BIGNUM(Scm_BignumComplement(b));
@@ -312,74 +322,80 @@ ScmInt64 Scm_BignumToSI64(ScmBignum *b, int clamphi, int clamplo)
         }
     }
     return r;
-  err:
-    Scm_Error("argument out of range: %S", SCM_OBJ(b));
-    return r; /*dummy*/
 #else  /*!SCM_EMULATE_INT64*/
     int64_t r = 0;
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (b->sign > 0) {
-        if (b->size > 2 || b->values[1] > LONG_MAX) {
-            if (!clamphi) goto err;
+        if (b->size == 1) {
+            r = b->values[0];
+        } else if (b->size > 2 || b->values[1] > LONG_MAX) {
+            if (!(clamp & SCM_CLAMP_HI)) goto err;
             SCM_SET_INT64_MAX(r);
         } else {
             r = ((int64_t)b->values[1] << 32) + (uint64_t)b->values[0];
         }
     } else { /* b->sign < 0 */
-        if (b->size > 2
-            || ((b->values[1] > (u_long)LONG_MAX)
-                && (b->values[0] > 0))) {
-            if (!clamplo) goto err;
+        if (b->size == 1) {
+            r = -(int64_t)b->values[0];
+        } else if (b->size > 2 || (b->values[1] > LONG_MAX && b->values[0] > 0)) {
+            if (!(clamp&SCM_CLAMP_LO)) goto err;
             SCM_SET_INT64_MIN(r);
         } else {
-            b = SCM_BIGNUM(Scm_BignumComplement(b));
-            r = ((int64_t)b->values[1] << 32) + (uint64_t)b->values[0];
+            r = -(((int64_t)b->values[1] << 32) + (uint64_t)b->values[0]);
         }
     }
     return r;
-  err:
-    Scm_Error("argument out of range: %S", SCM_OBJ(b));
-    return 0; /*dummy*/
 #endif /*!SCM_EMULATE_INT64*/
+  err:
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) {
+        *oor = TRUE;
+    } else {
+        Scm_Error("argument out of range: %S", SCM_OBJ(b));
+    }
+    return r;
 }
 
-ScmUInt64 Scm_BignumToUI64(ScmBignum *b, int clamphi, int clamplo)
+ScmUInt64 Scm_BignumToUI64(ScmBignum *b, int clamp, int *oor)
 {
 #if SCM_EMULATE_INT64
     ScmInt64 r = {0, 0};
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (b->sign > 0) {
         if (b->size > 2) {
-            if (!clamphi) goto err;
+            if (!(clamp&SCM_CLAMP_HI)) goto err;
             SCM_SET_UINT64_MAX(r);
         } else {
             r.lo = b->values[0];
             if (b->size == 2) r.hi = b->values[1];
         }
     } else if (b->sign < 0) {
-        if (!clamplo) goto err;
+        if (!(clamp&SCM_CLAMP_LO)) goto err;
     }
     return r;
-  err:
-    Scm_Error("argument out of range: %S", SCM_OBJ(b));
-    return r; /*dummy*/
 #else  /*!SCM_EMULATE_INT64*/
     uint64_t r = 0;
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) *oor = FALSE;
     if (b->sign > 0) {
         if (b->size > 2) {
-            if (!clamphi) goto err;
-            r = SCM_SET_UINT64_MAX(r);
+            if (!(clamp&SCM_CLAMP_HI)) goto err;
+            SCM_SET_UINT64_MAX(r);
         } else if (b->size == 2) {
-            r = ((uint64_t)b->values[1] << 32) + (uint64_t)b->values[0];
+            r = (((uint64_t)b->values[1]) << 32) + (uint64_t)b->values[0];
         } else {
             r = (uint64_t)b->values[0];
         }
     } else { /* b->sign < 0 */
-        if (!clamplo) goto err;
+        if (!(clamp&SCM_CLAMP_LO)) goto err;
     }
     return r;
-  err:
-    Scm_Error("argument out of range: %S", SCM_OBJ(b));
-    return 0; /*dummy*/
 #endif /*!SCM_EMULATE_INT64*/
+  err:
+    if (clamp == SCM_CLAMP_NONE && oor != NULL) {
+        *oor = TRUE;
+    } else {
+        Scm_Error("argument out of range: %S", SCM_OBJ(b));
+    }
+    return r;
 }
 #endif /* SIZEOF_LONG == 4 */
 
@@ -681,51 +697,6 @@ ScmObj Scm_BignumSubSI(ScmBignum *bx, long y)
     return Scm_NormalizeBignum(bignum_add_si(bx, -y));
 }
 
-ScmObj Scm_BignumAddN(ScmBignum *bx, ScmObj args)
-{
-    ScmBignum *r = bx;
-    for (;SCM_PAIRP(args); args = SCM_CDR(args)) {
-        ScmObj v = SCM_CAR(args);
-        if (SCM_INTP(v)) {
-            r = bignum_add_si(r, SCM_INT_VALUE(v));
-            continue;
-        }
-        if (SCM_BIGNUMP(v)) {
-            r = bignum_add(r, SCM_BIGNUM(v));
-            continue;
-        }
-        if (SCM_FLONUMP(v) || SCM_COMPLEXP(v)) {
-            ScmObj z = Scm_MakeFlonum(Scm_BignumToDouble(r));
-            return Scm_Add(z, v, SCM_CDR(args));
-        }
-        /* Pass back to Scm_Add to deal with object-add hook */
-        return Scm_Add(Scm_NormalizeBignum(r), v, SCM_CDR(args));
-    }
-    return Scm_NormalizeBignum(r);
-}
-
-ScmObj Scm_BignumSubN(ScmBignum *bx, ScmObj args)
-{
-    ScmBignum *r = bx;
-    for (;SCM_PAIRP(args); args = SCM_CDR(args)) {
-        ScmObj v = SCM_CAR(args);
-        if (SCM_INTP(v)) {
-            r = bignum_add_si(r, -SCM_INT_VALUE(v));
-            continue;
-        }
-        if (SCM_BIGNUMP(v)) {
-            r = bignum_sub(r, SCM_BIGNUM(v));
-            continue;
-        }
-        if (SCM_FLONUMP(v) || SCM_COMPLEXP(v)) {
-            ScmObj z = Scm_MakeFlonum(Scm_BignumToDouble(r));
-            return Scm_Subtract(z, v, SCM_CDR(args));
-        }
-        Scm_Error("number expected, but got %S", v);
-    }
-    return Scm_NormalizeBignum(r);
-}
-
 /*-----------------------------------------------------------------------
  * Shifter
  */
@@ -875,28 +846,6 @@ ScmObj Scm_BignumMulSI(ScmBignum *bx, long y)
     return Scm_NormalizeBignum(br);
 }
 
-ScmObj Scm_BignumMulN(ScmBignum *bx, ScmObj args)
-{
-    ScmBignum *r = bx;
-    for (; SCM_PAIRP(args); args = SCM_CDR(args)) {
-        ScmObj v = SCM_CAR(args);
-        if (SCM_INTP(v)) {
-            r = bignum_mul_si(r, SCM_INT_VALUE(v));
-            continue;
-        }
-        if (SCM_BIGNUMP(v)) {
-            r = bignum_mul(r, SCM_BIGNUM(v));
-            continue;
-        }
-        if (SCM_FLONUMP(v) || SCM_COMPLEXP(v)) {
-            ScmObj f = Scm_MakeFlonum(Scm_BignumToDouble(r));
-            return Scm_Multiply(f, v, SCM_CDR(args));
-        }
-        Scm_Error("number expected, but got %S", v);
-    }
-    return Scm_NormalizeBignum(r);
-}
-
 /*-----------------------------------------------------------------------
  * Division
  */
@@ -1101,11 +1050,10 @@ ScmObj Scm_BignumAsh(ScmBignum *x, int cnt)
         } else {
             if (SCM_BIGNUM_SIGN(x) < 0) {
                 /* painful way */
-                ScmObj r = Scm_Quotient(Scm_Add(SCM_OBJ(x), SCM_MAKE_INT(1),
-                                                SCM_NIL),
+                ScmObj r = Scm_Quotient(Scm_Add(SCM_OBJ(x), SCM_MAKE_INT(1)),
                                         Scm_Ash(SCM_MAKE_INT(1), -cnt),
                                         NULL);
-                return Scm_Add(r, SCM_MAKE_INT(-1), SCM_NIL);
+                return Scm_Add(r, SCM_MAKE_INT(-1));
             } else {
                 ScmBignum *r = make_bignum(rsize);
                 return Scm_NormalizeBignum(bignum_rshift(r, x, -cnt));
