@@ -1,7 +1,7 @@
 /*
  * char.c - character and character set operations
  *
- *   Copyright (c) 2000-2004 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2007  Shiro Kawai  <shiro@acm.org>
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: char.c,v 1.40 2004/01/17 01:34:48 shirok Exp $
+ *  $Id: char.c,v 1.53 2007/04/18 02:24:09 shirok Exp $
  */
 
 #include <ctype.h>
@@ -67,7 +67,12 @@ int Scm_SupportedCharacterEncodingP(const char *encoding)
 {
     const char **cs = supportedCharacterEncodings;
     for (;*cs;cs++) {
-        if (strcmp(*cs, encoding) == 0) return TRUE;
+        const char *p = *cs;
+        const char *q = encoding;
+        for (;*p && *q; p++, q++) {
+            if (tolower(*p) != tolower(*q)) break;
+        }
+        if (*p == '\0' && *q == '\0') return TRUE;
     }
     return FALSE;
 }
@@ -78,7 +83,7 @@ int Scm_DigitToInt(ScmChar ch, int radix)
 {
     if (ch < '0') return -1;
     if (radix <= 10) {
-        if (ch <= '0' + radix) return (ch - '0');
+        if (ch < '0' + radix) return (ch - '0');
     } else {
         if (ch <= '9') return (ch - '0');
         if (ch < 'A') return -1;
@@ -119,7 +124,8 @@ ScmChar Scm_UcsToChar(int n)
     if (Scm_UcsToCharHook == NULL) {
         /* NB: we don't need mutex here, for the loading of gauche.charconv
            is serialized in Scm_Require. */
-        Scm_Require(SCM_MAKE_STR("gauche/charconv"));
+        Scm_Require(SCM_MAKE_STR("gauche/charconv"),
+                    SCM_LOAD_PROPAGATE_ERROR, NULL);
         if (Scm_UcsToCharHook == NULL) {
             Scm_Error("couldn't autoload gauche.charconv");
         }
@@ -141,7 +147,8 @@ int Scm_CharToUcs(ScmChar ch)
     if (Scm_CharToUcsHook == NULL) {
         /* NB: we don't need mutex here, for the loading of gauche.charconv
            is serialized in Scm_Require. */
-        Scm_Require(SCM_MAKE_STR("gauche/charconv"));
+        Scm_Require(SCM_MAKE_STR("gauche/charconv"),
+                    SCM_LOAD_PROPAGATE_ERROR, NULL);
         if (Scm_CharToUcsHook == NULL) {
             Scm_Error("couldn't autoload gauche.charconv");
         }
@@ -165,25 +172,10 @@ SCM_DEFINE_BUILTIN_CLASS(Scm_CharSetClass,
                          charset_print, charset_compare, NULL, NULL,
                          SCM_CLASS_DEFAULT_CPL);
 
-/* masks */
-#if SIZEOF_LONG == 4
-#define MASK_BIT_SHIFT  5
-#define MASK_BIT_MASK   0x1f
-#elif SIZEOF_LONG == 8
-#define MASK_BIT_SHIFT  6
-#define MASK_BIT_MASK   0x3f
-#elif SIZEOF_LONG == 16    /* maybe, in some future ... */
-#define MASK_BIT_SHIFT  7
-#define MASK_BIT_MASK   0x7f
-#else
-#error need to set SIZEOF_LONG
-#endif
+#define MASK_ISSET(cs, ch)  SCM_BITS_TEST(cs->small, ch)
+#define MASK_SET(cs, ch)    SCM_BITS_SET(cs->small, ch)
+#define MASK_RESET(cs, ch)  SCM_BITS_RESET(cs->small, ch)
 
-#define MASK_INDEX(ch)       ((ch) >> MASK_BIT_SHIFT)
-#define MASK_BIT(ch)         (1L << ((ch) & MASK_BIT_MASK))
-#define MASK_ISSET(cs, ch)   (!!(cs->mask[MASK_INDEX(ch)] & MASK_BIT(ch)))
-#define MASK_SET(cs, ch)     (cs->mask[MASK_INDEX(ch)] |= MASK_BIT(ch))
-#define MASK_RESET(cs, ch)   (cs->mask[MASK_INDEX(ch)] &= ~MASK_BIT(ch))
 
 /*----------------------------------------------------------------------
  * Printer
@@ -193,12 +185,7 @@ static void charset_print_ch(ScmPort *out, ScmChar ch)
     if (ch < 0x20 || ch == 0x7f) {
         Scm_Printf(out, "\\x%02x", ch);
     } else {
-        char chbuf[SCM_CHAR_MAX_BYTES];
-        int i;
-        SCM_CHAR_PUT(chbuf, ch);
-        for (i=0; i<SCM_CHAR_NBYTES(ch); i++) {
-            Scm_Printf(out, "%c", chbuf[i]);
-        }
+        Scm_Putc(ch, out);
     }
 }
 
@@ -206,10 +193,11 @@ static void charset_print(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
 {
     int prev, code;
     ScmCharSet *cs = SCM_CHARSET(obj);
-    struct ScmCharSetRange *r;
+    ScmTreeIter iter;
+    ScmDictEntry *e;
 
     Scm_Printf(out, "#[");
-    for (prev = -1, code = 0; code < SCM_CHARSET_MASK_CHARS; code++) {
+    for (prev = -1, code = 0; code < SCM_CHARSET_SMALL_CHARS; code++) {
         if (MASK_ISSET(cs, code) && prev < 0) {
             charset_print_ch(out, code);
             prev = code;
@@ -228,11 +216,13 @@ static void charset_print(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
             charset_print_ch(out, code-1);
         }
     }
-    for (r = cs->ranges; r; r = r->next) {
-        charset_print_ch(out, r->lo);
-        if (r->hi == r->lo) continue;
-        if (r->hi - r->lo > 2) Scm_Printf(out, "-");
-        charset_print_ch(out, r->hi);
+    Scm_TreeIterInit(&iter, &cs->large, NULL);
+    while ((e = Scm_TreeIterNext(&iter)) != NULL) {
+        charset_print_ch(out, (int)e->key);
+        if (e->value != e->key) {
+            if (e->value - e->key > 2) Scm_Printf(out, "-");
+            charset_print_ch(out, (int)e->value);
+        }
     }
     Scm_Printf(out, "]", obj);
 }
@@ -240,13 +230,20 @@ static void charset_print(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
 /*-----------------------------------------------------------------
  * Constructors
  */
+static int cmp(ScmTreeCore *tc, intptr_t a, intptr_t b)
+{
+    if (a > b) return 1;
+    if (a < b) return -1;
+    return 0;
+}
+
 static ScmCharSet *make_charset(void)
 {
     ScmCharSet *cs = SCM_NEW(ScmCharSet);
     int i;
     SCM_SET_CLASS(cs, SCM_CLASS_CHARSET);
-    for (i=0; i<SCM_CHARSET_MASK_SIZE; i++) cs->mask[i] = 0;
-    cs->ranges = NULL;
+    Scm_BitsFill(cs->small, 0, SCM_CHARSET_SMALL_CHARS, 0);
+    Scm_TreeCoreInit(&cs->large, cmp, NULL);
     return cs;
 }
 
@@ -255,24 +252,11 @@ ScmObj Scm_MakeEmptyCharSet(void)
     return SCM_OBJ(make_charset());
 }
 
-ScmObj Scm_CopyCharSet(ScmCharSet *src)
+ScmObj Scm_CharSetCopy(ScmCharSet *src)
 {
     ScmCharSet *dst = make_charset();
-    struct ScmCharSetRange *rs, *rd = dst->ranges;
-    int i;
-    
-    for (i=0; i<SCM_CHARSET_MASK_SIZE; i++) dst->mask[i] = src->mask[i];
-    for (rs = src->ranges; rs; rs = rs->next) {
-        if (rd == NULL) {
-            rd = dst->ranges = SCM_NEW(struct ScmCharSetRange);
-        } else {
-            rd->next = SCM_NEW(struct ScmCharSetRange);
-            rd = rd->next;
-        }
-        rd->lo = rs->lo;
-        rd->hi = rs->hi;
-    }
-    if (rd) rd->next = NULL;
+    Scm_BitsCopyX(dst->small, 0, src->small, 0, SCM_CHARSET_SMALL_CHARS);
+    Scm_TreeCoreCopy(&dst->large, &src->large);
     return SCM_OBJ(dst);
 }
 
@@ -337,38 +321,56 @@ ScmChar Scm_ReadXdigitsFromPort(ScmPort *port, int ndigits,
  */
 static int charset_compare(ScmObj x, ScmObj y, int equalp)
 {
-    return 1;                   /* for now */
+    ScmCharSet *xx = SCM_CHARSET(x);
+    ScmCharSet *yy = SCM_CHARSET(y);
+    
+    if (equalp) {
+        return (Scm_CharSetEq(xx, yy)? 0 : 1);
+    } else {
+        if (Scm_CharSetEq(xx, yy)) return 0;
+        if (Scm_CharSetLE(xx, yy)) return -1;
+        if (Scm_CharSetLE(yy, xx)) return 1;
+        Scm_Error("cannot compare char-sets: %S vs %S", x, y);
+        return 0;               /* dummy */
+    }
 }
 
 int Scm_CharSetEq(ScmCharSet *x, ScmCharSet *y)
 {
-    int i;
-    struct ScmCharSetRange *rx, *ry;
-    for (i=0; i<SCM_CHARSET_MASK_SIZE; i++)
-        if (x->mask[i] != y->mask[i]) return FALSE;
-    for (rx=x->ranges, ry=y->ranges; rx && ry; rx=rx->next, ry=ry->next) {
-        if (rx->lo != ry->lo || rx->hi != ry->hi) return FALSE;
-    }
-    if (rx || ry) return FALSE;
+    if (!Scm_BitsEqual(x->small, y->small, 0, SCM_CHARSET_SMALL_CHARS))
+        return FALSE;
+    if (!Scm_TreeCoreEq(&x->large, &y->large))
+        return FALSE;
     return TRUE;
 }
 
 /* whether x <= y */
 int Scm_CharSetLE(ScmCharSet *x, ScmCharSet *y)
 {
-    int i;
-    struct ScmCharSetRange *rx, *ry;
-    for (i=0; i<SCM_CHARSET_MASK_SIZE; i++)
-        if ((x->mask[i] | y->mask[i]) != y->mask[i]) return FALSE;
-    rx = x->ranges;
-    ry = y->ranges;
-    while (rx && ry) {
-        if (rx->lo < ry->lo) return FALSE;
-        if (rx->lo > ry->hi) { ry = ry->next; continue; }
-        if (rx->hi > ry->hi) return FALSE;
-        rx = rx->next;
+    ScmTreeIter xi;
+    ScmDictEntry *xe, *ye, *yl, *yh;
+    if (!Scm_BitsIncludes(y->small, x->small, 0, SCM_CHARSET_SMALL_CHARS))
+        return FALSE;
+    /* For each range of X, check if it is fully contained by a range of Y.
+     *
+     * Case 1:
+     *    xk<---------->xv
+     *    yk<----------------->yv
+     * Case 2:
+     *         xk<---------->xv
+     *    yk<------------------>yv
+     */
+    Scm_TreeIterInit(&xi, &x->large, NULL);
+    for (xe = Scm_TreeIterNext(&xi); xe; xe = Scm_TreeIterNext(&xi)) {
+        ye = Scm_TreeCoreClosestEntries(&y->large, xe->key, &yl, &yh);
+        if (ye) {               /* case 1 */
+            if (ye->value < xe->value) return FALSE;
+        } else if (yl) {        /* case 2 */
+            if (yl->value < xe->value) return FALSE;
+        } else {
+            return FALSE;
+        }
     }
-    if (rx) return FALSE;
     return TRUE;
 }
 
@@ -376,94 +378,63 @@ int Scm_CharSetLE(ScmCharSet *x, ScmCharSet *y)
  * Modification
  */
 
-static struct ScmCharSetRange *newrange(int lo, int hi,
-                                        struct ScmCharSetRange *next)
-{
-    struct ScmCharSetRange *n = SCM_NEW(struct ScmCharSetRange);
-    n->next = next;
-    n->lo = lo;
-    n->hi = hi;
-    return n;
-}
-
 ScmObj Scm_CharSetAddRange(ScmCharSet *cs, ScmChar from, ScmChar to)
 {
-    int i;
-    struct ScmCharSetRange *lo, *lop, *hi;
+    ScmDictEntry *e, *lo, *hi;
+    int lbound;
     
     if (to < from) return SCM_OBJ(cs);
-    if (from < SCM_CHARSET_MASK_CHARS) {
-        if (to < SCM_CHARSET_MASK_CHARS) {
-            for (i=from; i<=to; i++) MASK_SET(cs, i);
+    if (from < SCM_CHARSET_SMALL_CHARS) {
+        if (to < SCM_CHARSET_SMALL_CHARS) {
+            Scm_BitsFill(cs->small, (int)from, (int)to+1, TRUE);
             return SCM_OBJ(cs);
         }
-        for (i=from; i<SCM_CHARSET_MASK_CHARS; i++)  MASK_SET(cs, i);
-        from = SCM_CHARSET_MASK_CHARS;
+        Scm_BitsFill(cs->small, (int)from, SCM_CHARSET_SMALL_CHARS, TRUE);
+        from = SCM_CHARSET_SMALL_CHARS;
     }
-    if (cs->ranges == NULL) {
-        cs->ranges = newrange(from, to, NULL);
-        return SCM_OBJ(cs);
-    }
-    /* Add range.  Ranges are chained from lower character code to higher,
-       without any overlap. */
-    /* First, we scan the ranges so that we'll get...
-        - if FROM is in a range, lo points to it.
-        - if FROM is out of any ranges, lo points to the closest range that
-          is higher than FROM.
-        - if TO is in a range, hi points to the range.
-        - if TO is out of any ranges, hi points to the closest range that
-          is higher than TO. */
-    for (lop = NULL, lo = cs->ranges; lo; lop = lo, lo = lo->next) {
-        if (from <= lo->hi+1) break;
-    }
-    if (!lo) {
-        lop->next = newrange(from, to, NULL);
-        return SCM_OBJ(cs);
-    }
-    for (hi = lo; hi; hi = hi->next) {
-        if (to <= hi->hi) break;
-    }
-    /* Then we insert, extend and/or merge the ranges accordingly. */
-    if (from < lo->lo) { /* FROM extends the LO */
-        if (lo == hi) {
-            if (to < hi->lo-1) {
-                if (lop == NULL) cs->ranges = newrange(from, to, lo);
-                else             lop->next = newrange(from, to, lo);
-            } else {
-                lo->lo = from;
-            }
-        } else if (hi == NULL || to < hi->lo-1) {
-            lo->lo = from;
-            lo->hi = to;
-            lo->next = hi;
+
+    /* Let e have the lower bound. */
+    e = Scm_TreeCoreClosestEntries(&cs->large, from, &lo, &hi);
+    if (!e) {
+        if (!lo || lo->value < from-1) {
+            e = Scm_TreeCoreSearch(&cs->large, from, SCM_DICT_CREATE);
         } else {
-            lo->lo = from;
-            lo->hi = hi->hi;
-            lo->next = hi->next;
-        }
-    } else { /* FROM included in LO */
-        if (lo != hi) {
-            if (hi == NULL || to < hi->lo-1) {
-                lo->hi = to;
-                lo->next = hi;
-            } else {
-                lo->hi = hi->hi;
-                lo->next = hi->next;
-            }
+            e = lo;
         }
     }
-    /* WRITE ME */
+    /* Set up the upper bound.
+       NB: if e is a new entry, e->value is 0. */
+    if (e->value >= to) return SCM_OBJ(cs);
+
+    hi = e;
+    while ((hi = Scm_TreeCoreNextEntry(&cs->large, hi->key)) != NULL) {
+        if (hi->key > to+1) {
+            e->value = to;
+            return SCM_OBJ(cs);
+        }
+        Scm_TreeCoreSearch(&cs->large, hi->key, SCM_DICT_DELETE);
+        if (hi->value > to) {
+            e->value = hi->value;
+            return SCM_OBJ(cs);
+        }
+    }
+    e->value = to;
     return SCM_OBJ(cs);
 }
 
 ScmObj Scm_CharSetAdd(ScmCharSet *dst, ScmCharSet *src)
 {
     int i;
-    struct ScmCharSetRange *r;
-    for (i=0; i<SCM_CHARSET_MASK_SIZE; i++)
-        dst->mask[i] |= src->mask[i];
-    for (r = src->ranges; r; r = r->next) {
-        Scm_CharSetAddRange(dst, r->lo, r->hi);
+    ScmTreeIter iter;
+    ScmDictEntry *e;
+
+    if (dst == src) return SCM_OBJ(dst);  /* precaution */
+    
+    Scm_BitsOperate(dst->small, SCM_BIT_IOR, dst->small, src->small,
+                    0, SCM_CHARSET_SMALL_CHARS);
+    Scm_TreeIterInit(&iter, &src->large, NULL);
+    while ((e = Scm_TreeIterNext(&iter)) != NULL) {
+        Scm_CharSetAddRange(dst, e->key, e->value);
     }
     return SCM_OBJ(dst);
 }
@@ -471,23 +442,23 @@ ScmObj Scm_CharSetAdd(ScmCharSet *dst, ScmCharSet *src)
 ScmObj Scm_CharSetComplement(ScmCharSet *cs)
 {
     int i, last;
-    struct ScmCharSetRange *r, *p;
-    for (i=0; i<SCM_CHARSET_MASK_SIZE; i++)
-        cs->mask[i] = ~cs->mask[i];
-    last = SCM_CHARSET_MASK_CHARS;
-    for (p = NULL, r = cs->ranges; r; p = r, r = r->next) {
-        int hi = r->hi+1;
-        if (r->lo != SCM_CHARSET_MASK_CHARS) {
-            r->hi = r->lo - 1;
-            r->lo = last;
-        } else {
-            cs->ranges = r->next;
+    ScmDictEntry *e, *n;
+
+    Scm_BitsOperate(cs->small, SCM_BIT_NOT1, cs->small, NULL,
+                    0, SCM_CHARSET_SMALL_CHARS);
+    last = SCM_CHARSET_SMALL_CHARS;
+    /* we can't use treeiter, since we modify the tree while traversing it. */
+    while ((e = Scm_TreeCoreNextEntry(&cs->large, last)) != NULL) {
+        Scm_TreeCoreSearch(&cs->large, e->key, SCM_DICT_DELETE);
+        if (last < e->key-1) {
+            n = Scm_TreeCoreSearch(&cs->large, last, SCM_DICT_CREATE);
+            n->value = e->key-1;
         }
-        last = hi;
+        last = e->value+1;
     }
     if (last < SCM_CHAR_MAX) {
-        if (!p) cs->ranges = newrange(last, SCM_CHAR_MAX, NULL);
-        else    p->next = newrange(last, SCM_CHAR_MAX, NULL);
+        n = Scm_TreeCoreSearch(&cs->large, last, SCM_DICT_CREATE);
+        n->value = SCM_CHAR_MAX;
     }
     return SCM_OBJ(cs);
 }
@@ -513,13 +484,12 @@ ScmObj Scm_CharSetCaseFold(ScmCharSet *cs)
 int Scm_CharSetContains(ScmCharSet *cs, ScmChar c)
 {
     if (c < 0) return FALSE;
-    if (c < SCM_CHARSET_MASK_CHARS) return MASK_ISSET(cs, c);
+    if (c < SCM_CHARSET_SMALL_CHARS) return MASK_ISSET(cs, c);
     else {
-        struct ScmCharSetRange *r;
-        for (r = cs->ranges; r; r = r->next) {
-            if (r->lo <= c && c <= r->hi) return TRUE;
-        }
-        return FALSE;
+        ScmDictEntry *e, *l, *h;
+        e = Scm_TreeCoreClosestEntries(&cs->large, (int)c, &l, &h);
+        if (e || (l && l->value >= c)) return TRUE;
+        else return FALSE;
     }
 }
 
@@ -532,9 +502,10 @@ ScmObj Scm_CharSetRanges(ScmCharSet *cs)
 {
     ScmObj h = SCM_NIL, t = SCM_NIL, cell;
     int ind, begin = 0, prev = FALSE;
-    struct ScmCharSetRange *r;
-    
-    for (ind = 0; ind < SCM_CHARSET_MASK_CHARS; ind++) {
+    ScmTreeIter iter;
+    ScmDictEntry *e;
+
+    for (ind = 0; ind < SCM_CHARSET_SMALL_CHARS; ind++) {
         int bit = MASK_ISSET(cs, ind);
         if (!prev && bit) begin = ind;
         if (prev && !bit) {
@@ -544,21 +515,12 @@ ScmObj Scm_CharSetRanges(ScmCharSet *cs)
         prev = bit;
     }
     if (prev) {
-        if (!cs->ranges || cs->ranges->lo != SCM_CHARSET_MASK_CHARS) {
-            cell = Scm_Cons(SCM_MAKE_INT(begin),
-                            SCM_MAKE_INT(SCM_CHARSET_MASK_CHARS-1));
-            SCM_APPEND1(h, t, cell);
-            r = cs->ranges;
-        } else {
-            cell = Scm_Cons(SCM_MAKE_INT(begin), SCM_MAKE_INT(cs->ranges->hi));
-            SCM_APPEND1(h, t, cell);
-            r = cs->ranges->next;
-        }
-    } else {
-        r = cs->ranges;
+        cell = Scm_Cons(SCM_MAKE_INT(begin), SCM_MAKE_INT(ind-1));
+        SCM_APPEND1(h, t, cell);
     }
-    for (; r; r = r->next) {
-        cell = Scm_Cons(SCM_MAKE_INT(r->lo), SCM_MAKE_INT(r->hi));
+    Scm_TreeIterInit(&iter, &cs->large, NULL);
+    while ((e = Scm_TreeIterNext(&iter)) != NULL) {
+        cell = Scm_Cons(SCM_MAKE_INT(e->key), SCM_MAKE_INT(e->value));
         SCM_APPEND1(h, t, cell);
     }
     return h;
@@ -570,11 +532,15 @@ void Scm_CharSetDump(ScmCharSet *cs, ScmPort *port)
     int i;
     struct ScmCharSetRange *r;
     Scm_Printf(port, "CharSet %p\nmask:", cs);
-    for (i=0; i<SCM_CHARSET_MASK_SIZE; i++)
-        Scm_Printf(port, "[%08x]", cs->mask[i]);
+    for (i=0; i<SCM_BITS_NUM_WORDS(SCM_CHARSET_SMALL_CHARS); i++) {
+#if SIZEOF_LONG == 4
+        Scm_Printf(port, "[%08lx]", cs->small[i]);
+#else
+        Scm_Printf(port, "[%016lx]", cs->small[i]);
+#endif
+    }
     Scm_Printf(port, "\nranges:");
-    for (r=cs->ranges; r; r=r->next)
-        Scm_Printf(port, "(%d-%d)", r->lo, r->hi);
+    Scm_TreeCoreDump(&cs->large, port);
     Scm_Printf(port, "\n");
 }
 #endif /* SCM_DEBUG_HELPER */
@@ -723,7 +689,7 @@ ScmObj Scm_CharSetRead(ScmPort *input, int *complement_p,
                 goto ordchar;
             }
             if (moreset_complement) {
-                moreset = Scm_CharSetComplement(SCM_CHARSET(Scm_CopyCharSet(SCM_CHARSET(moreset))));
+                moreset = Scm_CharSetComplement(SCM_CHARSET(Scm_CharSetCopy(SCM_CHARSET(moreset))));
             }
             Scm_CharSetAdd(set, SCM_CHARSET(moreset));
             continue;
@@ -836,7 +802,7 @@ static void install_charsets(void)
     for (i = 0; i < SCM_CHARSET_NUM_PREDEFINED_SETS; i++) {
         CS(i) = SCM_CHARSET(Scm_MakeEmptyCharSet());
     }
-    for (code = 0; code < SCM_CHARSET_MASK_CHARS; code++) {
+    for (code = 0; code < SCM_CHARSET_SMALL_CHARS; code++) {
         if (isalnum(code)) MASK_SET(CS(SCM_CHARSET_ALNUM), code);
         if (isalpha(code)) MASK_SET(CS(SCM_CHARSET_ALPHA), code);
         if (iscntrl(code)) MASK_SET(CS(SCM_CHARSET_CNTRL), code);
@@ -873,4 +839,15 @@ ScmObj Scm_GetStandardCharSet(int id)
 void Scm__InitChar(void)
 {
     SCM_INTERNAL_MUTEX_INIT(predef_charsets_mutex);
+
+    /* Expose internal charset */
+#if defined(GAUCHE_CHAR_ENCODING_EUC_JP)
+    Scm_AddFeature("gauche.ces.eucjp", NULL);
+#elif defined(GAUCHE_CHAR_ENCODING_SJIS)
+    Scm_AddFeature("gauche.ces.sjis", NULL);
+#elif defined(GAUCHE_CHAR_ENCODING_UTF_8)
+    Scm_AddFeature("gauche.ces.utf8", NULL);
+#else
+    Scm_AddFeature("gauche.ces.none", NULL);
+#endif
 }
