@@ -1,7 +1,7 @@
 /*
  * error.c - error handling
  *
- *   Copyright (c) 2000-2006 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2007  Shiro Kawai  <shiro@acm.org>
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: error.c,v 1.67 2006/01/23 07:06:32 shirok Exp $
+ *  $Id: error.c,v 1.77 2007/04/12 03:26:55 shirok Exp $
  */
 
 #include <errno.h>
@@ -269,7 +269,6 @@ static void readerror_dummy_set(ScmReadError *obj, ScmObj val)
     /* nothing */
 }
 
-
 static ScmObj porterror_port_get(ScmPortError *obj)
 {
     return obj->port? SCM_OBJ(obj->port) : SCM_FALSE;
@@ -528,11 +527,9 @@ void Scm_Error(const char *msg, ...)
     SCM_VM_RUNTIME_FLAG_SET(vm, SCM_ERROR_BEING_HANDLED);
     
     SCM_UNWIND_PROTECT {
-        ScmObj ostr = Scm_MakeOutputStringPort(TRUE);
         va_start(args, msg);
-        Scm_Vprintf(SCM_PORT(ostr), msg, args, TRUE);
+        e = Scm_MakeError(Scm_Vsprintf(msg, args, TRUE));
         va_end(args);
-        e = Scm_MakeError(Scm_GetOutputString(SCM_PORT(ostr)));
     }
     SCM_WHEN_ERROR {
         /* TODO: should check continuation? */
@@ -591,7 +588,7 @@ void Scm_SysError(const char *msg, ...)
         va_end(args);
         SCM_PUTZ(": ", -1, ostr);
         SCM_PUTS(syserr, ostr);
-        e = Scm_MakeSystemError(Scm_GetOutputString(SCM_PORT(ostr)), en);
+        e = Scm_MakeSystemError(Scm_GetOutputString(SCM_PORT(ostr), 0), en);
     }
     SCM_WHEN_ERROR {
         /* TODO: should check continuation */
@@ -603,7 +600,18 @@ void Scm_SysError(const char *msg, ...)
 }
 
 /*
- * A convenience function to throw port-relates errors.
+ * A convenience function to raise argument-type-violation errors.
+ * Right now it raises <error>, but it'll be changed once we adopt R6RS
+ * (The draft R6RS defines 'contract violation' in such cases).
+ */
+void Scm_TypeError(const char *what, const char *expected, ScmObj got)
+{
+    Scm_Error("%s expected for %s, but got %S", expected, what, got);
+}
+
+
+/*
+ * A convenience function to raise port-relates errors.
  * It creates either one of <port-error>, <io-read-error>,
  * <io-write-error>, <io-closed-error>, or <io-unit-error>,
  * depending on the 'reason' argument being 
@@ -629,7 +637,7 @@ void Scm_PortError(ScmPort *port, int reason, const char *msg, ...)
             SCM_PUTZ(": ", -1, ostr);
             SCM_PUTS(syserr, ostr);
         }
-        smsg = Scm_GetOutputString(SCM_PORT(ostr));
+        smsg = Scm_GetOutputString(SCM_PORT(ostr), 0);
 
         switch (reason) {
         case SCM_PORT_ERROR_INPUT:
@@ -671,12 +679,10 @@ void Scm_PortError(ScmPort *port, int reason, const char *msg, ...)
 void Scm_Warn(const char *msg, ...)
 {
     va_list args;
-    ScmObj ostr = Scm_MakeOutputStringPort(TRUE);
     va_start(args, msg);
-    Scm_Vprintf(SCM_PORT(ostr), msg, args, TRUE);
-    va_end(args);
-    Scm_Printf(SCM_CURERR, "WARNING: %A\n", Scm_GetOutputString(SCM_PORT(ostr)));
+    Scm_Printf(SCM_CURERR, "WARNING: %A\n", Scm_Vsprintf(msg, args, TRUE));
     Scm_Flush(SCM_CURERR);
+    va_end(args);
 }
 
 /* format & warn */
@@ -684,7 +690,8 @@ void Scm_FWarn(ScmString *fmt, ScmObj args)
 {
     ScmObj ostr = Scm_MakeOutputStringPort(TRUE);
     Scm_Format(SCM_PORT(ostr), fmt, args, TRUE);
-    Scm_Printf(SCM_CURERR, "WARNING: %A\n", Scm_GetOutputString(SCM_PORT(ostr)));
+    Scm_Printf(SCM_CURERR, "WARNING: %A\n",
+               Scm_GetOutputString(SCM_PORT(ostr), 0));
     Scm_Flush(SCM_CURERR);
 }
 
@@ -721,7 +728,6 @@ ScmObj Scm_Raise(ScmObj condition)
 
 ScmObj Scm_RaiseCondition(ScmObj condition_type, ...)
 {
-    ScmVM *vm = Scm_VM();
     ScmObj argh = SCM_NIL, argt = SCM_NIL;
     va_list ap;
 
@@ -741,7 +747,7 @@ ScmObj Scm_RaiseCondition(ScmObj condition_type, ...)
             ScmObj ostr = Scm_MakeOutputStringPort(TRUE);
             Scm_Vprintf(SCM_PORT(ostr), msg, ap, TRUE);
             SCM_APPEND1(argh, argt, SCM_MAKE_KEYWORD("message"));
-            SCM_APPEND1(argh, argt, Scm_GetOutputString(SCM_PORT(ostr)));
+            SCM_APPEND1(argh, argt, Scm_GetOutputString(SCM_PORT(ostr), 0));
             break;
         } else {
             ScmObj arg = va_arg(ap, ScmObj);
@@ -750,7 +756,7 @@ ScmObj Scm_RaiseCondition(ScmObj condition_type, ...)
         }
     }
     va_end(ap);
-    return Scm_Apply(SCM_SYMBOL_VALUE("gauche", "error"), argh);
+    return Scm_ApplyRec(SCM_SYMBOL_VALUE("gauche", "error"), argh);
 }
 
 /*
@@ -853,7 +859,7 @@ static void Scm_PrintDefaultErrorHeading(ScmObj e, ScmPort *out)
             Scm_Printf(out, "*** %s\n", heading);
         }
     } else {
-        Scm_Printf(out, "*** ERROR: unhandled exeption: %S\n", e);
+        Scm_Printf(out, "*** ERROR: unhandled exception: %S\n", e);
     }
 }
 
@@ -892,7 +898,7 @@ void Scm_ReportError(ScmObj e)
     SCM_VM_RUNTIME_FLAG_SET(vm, SCM_ERROR_BEING_REPORTED);
     SCM_UNWIND_PROTECT {
         if (SCM_PROCEDUREP(vm->defaultEscapeHandler)) {
-            Scm_Apply(vm->defaultEscapeHandler, SCM_LIST1(e));
+            Scm_ApplyRec(vm->defaultEscapeHandler, SCM_LIST1(e));
         } else {
             report_error_inner(vm, e);
         }
