@@ -1,7 +1,7 @@
 /*
  * vm.h - Virtual machine
  *
- *   Copyright (c) 2000-2004 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2007  Shiro Kawai  <shiro@acm.org>
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: vm.h,v 1.90 2004/07/15 23:16:13 shirok Exp $
+ *  $Id: vm.h,v 1.114 2007/03/22 11:20:36 shirok Exp $
  */
 
 #ifndef GAUCHE_VM_H
@@ -42,10 +42,22 @@
 /* Maximum # of values allowed for multiple value return */
 #define SCM_VM_MAX_VALUES      20
 
-/* Signal queue size */
-#define SCM_VM_SIGQ_SIZE       32
+/* Finalizer queue size */
+#define SCM_VM_FINQ_SIZE       32
 
-#define SCM_PCTYPE ScmObj
+/* QueueNotEmpty flag */
+#define SCM_VM_SIGQ_MASK       1
+#define SCM_VM_FINQ_MASK       2
+
+#define SCM_PCTYPE ScmWord*
+
+#if defined(ITIMER_PROF) && defined(SIGPROF)
+/* define if you want to use profiler */
+#define GAUCHE_PROFILE
+#endif /* defined(ITIMER_PROF) && defined(SIGPROF) */
+
+/* Actual structure is defined in code.h */
+typedef struct ScmCompiledCodeRec ScmCompiledCode;
 
 /*
  * Environment frame
@@ -65,8 +77,8 @@
 
 typedef struct ScmEnvFrameRec {
     struct ScmEnvFrameRec *up;  /* static link */
-    ScmObj info;                /* source code info for debug */
-    int size;                   /* size of the frame (excluding header) */
+    ScmObj info;                /* reserved */
+    ScmWord size;               /* size of the frame (excluding header) */
 } ScmEnvFrame;
 
 #define ENV_HDR_SIZE   3        /* envframe header size */
@@ -74,8 +86,6 @@ typedef struct ScmEnvFrameRec {
 #define ENV_FP(env)        (((ScmObj*)(env))-((env)->size))
 #define ENV_DATA(env, num) (*(((ScmObj*)(env))-(num)-1))
 
-SCM_EXTERN ScmEnvFrame *Scm_GetCurrentEnv(void);
-
 /*
  * Continuation frame
  *
@@ -89,13 +99,37 @@ typedef struct ScmContFrameRec {
     ScmObj *argp;                 /* saved argument pointer */
     int size;                     /* size of argument frame */
     SCM_PCTYPE pc;                /* next PC */
-    ScmObj info;                  /* debug info */
+    ScmCompiledCode *base;        /* base register value */
 } ScmContFrame;
 
 #define CONT_FRAME_SIZE  (sizeof(ScmContFrame)/sizeof(ScmObj))
 
 SCM_EXTERN void Scm_CallCC(ScmObj body);
 
+/*
+ * Syntactic closure
+ *
+ *   Syntactic closure encapsulates compile-time environment for
+ *   hygienic macro expansion.
+ *   See Bawden & Rees, Syntactic Closures, MIT AI Memo 1049, June 1988.
+ */
+
+typedef struct ScmSyntacticClosureRec {
+    ScmObj env;                 /* compile-time environment */
+    ScmObj literals;            /* literal symbols */
+    ScmObj expr;                /* expression */
+} ScmSyntacticClosure;
+
+SCM_CLASS_DECL(Scm_SyntacticClosureClass);
+#define SCM_CLASS_SYNTACTIC_CLOSURE   (&Scm_SyntacticClosureClass)
+
+#define SCM_SYNTACTIC_CLOSURE(obj)   ((ScmSyntacticClosure*)(obj))
+#define SCM_SYNTACTIC_CLOSURE_P(obj) SCM_XTYPEP(obj, SCM_CLASS_SYNTACTIC_CLOSURE)
+
+SCM_EXTERN ScmObj Scm_MakeSyntacticClosure(ScmObj env,
+                                           ScmObj literals,
+                                           ScmObj expr);
+
 /*
  * Identifier
  *
@@ -103,6 +137,9 @@ SCM_EXTERN void Scm_CallCC(ScmObj body);
  *   object is used in hygienic macro expansion (see macro.c), and
  *   also used as a placeholder in a global variable reference/assignment
  *   (see compile.c).
+ *
+ *   NB: Identifier's API and usage will likely be changed in future.
+ *   It shouldn't be used directly from applications.
  */
 
 typedef struct ScmIdentifierRec {
@@ -118,7 +155,8 @@ SCM_CLASS_DECL(Scm_IdentifierClass);
 #define SCM_IDENTIFIER(obj)     ((ScmIdentifier*)(obj))
 #define SCM_IDENTIFIERP(obj)    SCM_XTYPEP(obj, SCM_CLASS_IDENTIFIER)
 
-SCM_EXTERN ScmObj Scm_MakeIdentifier(ScmSymbol *name, ScmObj env);
+SCM_EXTERN ScmObj Scm_MakeIdentifier(ScmSymbol *name, ScmModule *mod,
+                                     ScmObj env);
 SCM_EXTERN ScmObj Scm_CopyIdentifier(ScmIdentifier *id);
 SCM_EXTERN int    Scm_IdentifierBindingEqv(ScmIdentifier *id, ScmSymbol *sym,
 					   ScmObj env);
@@ -140,12 +178,43 @@ typedef struct ScmCStackRec {
 /*
  * Escape point
  *
- *  EscapePoint structure keeps certain point of continuation chain
+ *  EscapePoint (EP) structure keeps certain point of continuation chain
  *  where control can be transferred.   This structure is used for
  *  saved continuations, as well as error handlers.
+ *
+ *  Normally EPs forms a single list, linked by the prev pointer.
+ *  vm->escapePoint points a 'current' EP, whose ehandler is the current
+ *  error handler.
+ *
+ *  However, this simple structure does not work in all cases.
+ *  When an error is signalled, we pop EP before executing the error
+ *  handler so that an error raised within the error handler will be
+ *  handled by the handler of outer EP.
+ *
+ *    Suppose the current EP is EP0.
+ *
+ *    (with-error-handler    ;; <- (1)this installs EP1. EP1->cont captures
+ *                           ;;        one-shot continuation of this expr.
+ *       (lambda (e) ...)    ;; <- (3)this is executed while EP0 is current
+ *      (lambda () ...)      ;; <- (2)this is executed while EP1 is current
+ *
+ *  If the error handler returns, we pass its result to the continuation of
+ *  with-error-handler, which is kept in EP1->cont.  The problem arises
+ *  if a stack overflow occurs within the error handler, continuation frames
+ *  in the stack are relocated to the heap, but EP1->cont isn't updated
+ *  since it is out of vm->escapePoint chain.
+ *
+ *  'Floating' pointer is used to catch such case.  When an EP is popped
+ *  before an error handler is ivoked, EP0's floating pointer is set to
+ *  point EP1.  When a new EP is pushed, it inherits the previous EP's
+ *  floating pointer.  With this scheme, active floating EPs are always
+ *  reachable from vm->esapePoint->floating chain.  (NB: the chain length
+ *  can be more than 1, if with-error-handler is used within an error handler
+ *  and an error is signalled in its body.
  */
 typedef struct ScmEscapePointRec {
     struct ScmEscapePointRec *prev;
+    struct ScmEscapePointRec *floating;
     ScmObj ehandler;            /* handler closure */
     ScmContFrame *cont;         /* saved continuation */
     ScmObj handlers;            /* saved dynamic handler chain */
@@ -155,8 +224,28 @@ typedef struct ScmEscapePointRec {
                                    when this ep is captured.  The flag status
                                    should be restored when the control
                                    transferred to this escape point. */
+    int rewindBefore;           /* EXPERIMENTAL: if TRUE, dynamic handlers
+                                   are rewound after an exception is raised
+                                   and before the exception handler is called.
+                                   If FALSE, the exception handler is called
+                                   first, then the dynamic handlers are 
+                                   rewound.   SRFI-18 model and legacy 
+                                   with-error-handler uses the latter model,
+                                   but SRFI-34's guard needs the former model.
+                                */
 } ScmEscapePoint;
 
+/* Link management */
+#define SCM_VM_FLOATING_EP(vm) \
+    ((vm)->escapePoint? (vm)->escapePoint->floating : vm->escapePointFloating)
+#define SCM_VM_FLOATING_EP_SET(vm, ep)          \
+    do {                                        \
+        if ((vm)->escapePoint) {                \
+            (vm)->escapePoint->floating = (ep); \
+        } else {                                \
+            (vm)->escapePointFloating = (ep);   \
+        }                                       \
+    } while (0)
 
 /* Escape types */
 #define SCM_VM_ESCAPE_NONE   0
@@ -165,26 +254,74 @@ typedef struct ScmEscapePointRec {
 #define SCM_VM_ESCAPE_EXIT   3
 
 /*
- * Parameters
+ * Signal queue
  *
- *  Parameters keep thread-local state.   It is called 'fluids' in some
- *  Scheme implementations.  A thread inherits the parameters from its
- *  creator.   
+ *  Following the Unix tradition, gauche doesn't really queue the signals;
+ *  the C-level signal handler only records the fact that a specific
+ *  signal has been delivered.  VM loop examines the record and invoke
+ *  appropriate Scheme-level signal handlers.  Even if the same signal
+ *  arrives more than one time before VM loop checks the record, the handler
+ *  is invoked only once.
+ *
+ *  The C-level signal handler does count each signal until it is cleared
+ *  by VM loop.  If one kind of signal arraives more than a certain limit
+ *  (set/get by Scm_{Set|Get}SignalPendingLimit), Gauche thinks something
+ *  went wrong and bail out.   It is useful, for example, to interrupt
+ *  unresponsive program from the terminal.
  */
 
-typedef struct ScmVMParameterTableRec {
-    int numParameters;
-    int numAllocated;
-    ScmObj *vector;
-    int *ids;
-} ScmVMParameterTable;
+typedef struct ScmSignalQueueRec {
+    unsigned char sigcounts[NSIG];
+    ScmObj pending;        /* pending signal handlers */
+} ScmSignalQueue;
 
-SCM_EXTERN void Scm_ParameterTableInit(ScmVMParameterTable *table,
-                                       ScmVM *base);
+SCM_EXTERN void Scm_SignalQueueInit(ScmSignalQueue* q);
+SCM_EXTERN int  Scm_GetSignalPendingLimit(void);
+SCM_EXTERN void Scm_SetSignalPendingLimit(int val);
 
-SCM_EXTERN int Scm_MakeParameterSlot(ScmVM *vm, int *newid);
-SCM_EXTERN ScmObj Scm_ParameterRef(ScmVM *vm, int index, int id);
-SCM_EXTERN ScmObj Scm_ParameterSet(ScmVM *vm, int index, int id, ScmObj value);
+/* NB: Obsoleted macro, but kept for backward compatibility.
+   Better API should be provided in future. */
+#define SCM_SIGPENDING(vm) \
+    ((vm)->queueNotEmpty&SCM_VM_SIGQ_MASK)
+
+#define SCM_SIGCHECK(vm) \
+    do { if (vm->queueNotEmpty&SCM_VM_SIGQ_MASK) Scm_SigCheck(vm); } while (0)
+
+SCM_EXTERN void   Scm_SigCheck(ScmVM *vm);
+
+/*
+ * Finalizers
+ *
+ *  Finalizers are queued inside GC.  We disable automatic finalizer
+ *  invocation of GC and only set the flag on VM when finalizers are
+ *  queued.  VM loop check the flag and calls Scm_VMFinalizerRun()
+ *  to run the finalizers.  (If VM is not running, C library must
+ *  call it explicitly to run finalizers).
+ */
+
+SCM_EXTERN ScmObj Scm_VMFinalizerRun(ScmVM *vm);
+
+/*
+ * Statistics
+ *
+ *  Not much stats are collected yet, but will grow in future.
+ *  Stats collections are only active if SCM_COLLECT_VM_STATS 
+ *  runtime flag is TRUE.
+ *  Stats are collected per-VM (i.e. per-thread), but currently
+ *  we don't have an API to gather them.
+ */
+
+typedef struct ScmVMStatRec {
+    /* Stack overflow handler */
+    u_long     sovCount; /* # of stack overflow */
+    double     sovTime;  /* cumulated time of stack ov handling */
+
+    /* Load statistics chain */
+    ScmObj     loadStat;
+} ScmVMStat;
+
+/* The profiler structure is defined in prof.h */
+typedef struct ScmVMProfilerRec ScmVMProfiler;
 
 /*
  * VM structure
@@ -227,6 +364,7 @@ struct ScmVMRec {
                                    "C stack rewinding" below. */
     unsigned int runtimeFlags;  /* Runtime flags */
     unsigned int compilerFlags; /* Compiler flags */
+    unsigned int queueNotEmpty; /* Bitmask if sigq or finq is not empty */
 
     ScmPort *curin;             /* current input port */
     ScmPort *curout;            /* current output port */
@@ -234,8 +372,9 @@ struct ScmVMRec {
     ScmVMParameterTable parameters; /* parameter table */
 
     /* Registers */
-    SCM_PCTYPE pc;              /* Program pointer.  Points list of
-                                   instructions to be executed.              */
+    ScmCompiledCode *base;      /* Current executing closure's code packet. */
+    SCM_PCTYPE pc;              /* Program pointer.  Points into the code
+                                   vector. (base->code) */
     ScmEnvFrame *env;           /* Current environment.                      */
     ScmContFrame *cont;         /* Current continuation.                     */
     ScmObj *argp;               /* Current argument pointer.  Points
@@ -260,6 +399,10 @@ struct ScmVMRec {
                                    continuation).  used by system's default
                                    exception handler to escape from the error
                                    handlers. */
+    ScmEscapePoint *escapePointFloating;
+                                /* reverse link of escape point chain
+                                   to keep 'active' EPs.
+                                   See ScmEscapePoint definition above. */
     int escapeReason;           /* temporary storage to pass data across
                                    longjmp(). */
     void *escapeData[2];        /* ditto. */
@@ -271,21 +414,25 @@ struct ScmVMRec {
     ScmObj load_next;           /* list of the directories to be searched */
     ScmObj load_history;        /* history of the nested load */
     ScmObj load_port;           /* current port from which we are loading */
+    int    evalSituation;       /* eval situation (related to eval-when) */
 
     /* Signal information */
-    int sigQueue[SCM_VM_SIGQ_SIZE];/* Ring buffer for pending signals */
-    unsigned int sigQueueHead;  /* points to the queue head */
-    unsigned int sigQueueTail;  /* points to the queue tail */
-    unsigned int sigOverflow;   /* flag to indicate queue overflow */
-    ScmObj sigPending;          /* pending signal handlers */
+    ScmSignalQueue sigq;
     sigset_t sigMask;           /* current signal mask */
+
+    /* Statistics */
+    ScmVMStat stat;
+    int profilerRunning;
+    ScmVMProfiler *prof;
 };
 
-SCM_EXTERN ScmVM *Scm_NewVM(ScmVM *base, ScmModule *module, ScmObj name);
+SCM_EXTERN ScmVM *Scm_NewVM(ScmVM *proto, ScmObj name);
+SCM_EXTERN int    Scm_AttachVM(ScmVM *vm);
 SCM_EXTERN void   Scm_VMDump(ScmVM *vm);
-SCM_EXTERN void   Scm_VMDefaultExceptionHandler(ScmObj);
-SCM_EXTERN ScmObj Scm_VMGetSourceInfo(ScmObj program);
-SCM_EXTERN ScmObj Scm_VMGetBindInfo(ScmObj program);
+SCM_EXTERN void   Scm_VMDefaultExceptionHandler(ScmObj exc);
+SCM_EXTERN ScmObj Scm_VMThrowException(ScmVM *vm, ScmObj exc);
+SCM_EXTERN ScmObj Scm_VMGetSourceInfo(ScmCompiledCode *code, SCM_PCTYPE pc);
+SCM_EXTERN ScmObj Scm_VMGetBindInfo(ScmCompiledCode *code, SCM_PCTYPE pc);
 
 SCM_CLASS_DECL(Scm_VMClass);
 #define SCM_CLASS_VM              (&Scm_VMClass)
@@ -309,40 +456,13 @@ enum {
                                    terminated. */
 };
 
-/*
- * VM instructions
- */
-#define SCM_VM_INSN_TAG            0x0e
-
-#define SCM_VM_INSNP(obj)            ((SCM_WORD(obj)&0x0f) == SCM_VM_INSN_TAG)
-#define SCM_VM_INSN_CODE(obj)        ((SCM_WORD(obj)>>4)&0x0ff)
-#define SCM_VM_INSN_ARG(obj)         ((signed long)SCM_WORD(obj) >> 12)
-
-#define SCM_VM_INSN_ARG0(obj)        ((SCM_WORD(obj) >> 12) & 0x03ff)
-#define SCM_VM_INSN_ARG1(obj)        ((SCM_WORD(obj) >> 22) & 0x03ff)
-
-#define SCM_VM_INSN(code) \
-    SCM_OBJ((long)((code)<<4)|SCM_VM_INSN_TAG)
-#define SCM_VM_INSN1(code, arg) \
-    SCM_OBJ((long)((arg) << 12) | ((code) << 4) | SCM_VM_INSN_TAG)
-#define SCM_VM_INSN2(code, arg0, arg1) \
-    SCM_OBJ((long)((arg1) << 22) | ((arg0) << 12) | ((code) << 4) | SCM_VM_INSN_TAG)
-
-#define SCM_VM_INSN_ARG_MAX          ((1L<<(32-13))-1)
-#define SCM_VM_INSN_ARG_MIN          (-SCM_VM_INSN_ARG_MAX)
-#define SCM_VM_INSN_ARG_FITS(k) \
-    (((k)<=SCM_VM_INSN_ARG_MAX)&&((k)>=SCM_VM_INSN_ARG_MIN))
-
+/* Value of vm->evalSituation */
 enum {
-#define DEFINSN(sym, nam, nparams)  sym,
-#include "vminsn.h"
-#undef DEFINSN
-    SCM_VM_NUM_INSNS
+    SCM_VM_EXECUTING,           /* we're evaluating the form interactively. */
+    SCM_VM_LOADING,             /* we're loading the forms */
+    SCM_VM_COMPILING            /* we're batch-compiling the forms */
 };
 
-SCM_EXTERN void   Scm__VMInsnWrite(ScmObj insn, ScmPort *port,
-				   ScmWriteContext *ctx);
-SCM_EXTERN ScmObj Scm_VMInsnInspect(ScmObj obj);
 
 /*
  * C stack rewinding
@@ -403,6 +523,10 @@ SCM_EXTERN ScmObj Scm_VMInsnInspect(ScmObj obj);
        Scm_VM()->cstack = Scm_VM()->cstack->prev;       \
     } while (0)
 
+SCM_EXTERN long Scm_VMUnwindProtect(ScmVM *vm, ScmCStack *cstack);
+SCM_EXTERN void Scm_VMNextHandler(ScmVM *vm);
+SCM_EXTERN void Scm_VMRewindProtect(ScmVM *vm);
+
 /*
  * Runtime flags
  */
@@ -411,9 +535,13 @@ enum {
     SCM_ERROR_BEING_REPORTED = (1L<<1), /* we're in an error reporter */
     SCM_LOAD_VERBOSE         = (1L<<2), /* report loading files */
     SCM_CASE_FOLD            = (1L<<3), /* symbols are case insensitive */
-    SCM_LIMIT_MODULE_MUTATION = (1L<<4)  /* disable set! to modify the
-                                            global binding in the other
-                                            module */
+    SCM_LIMIT_MODULE_MUTATION = (1L<<4),/* disable set! to modify the
+                                           global binding in the other
+                                           module */
+    SCM_COLLECT_VM_STATS     = (1L<<5), /* enable statistics collection
+                                           (incurs runtime overhead) */
+    SCM_COLLECT_LOAD_STATS   = (1L<<6)  /* log the stats of file load
+                                           timings (incurs runtime overhead) */
 };
 
 #define SCM_VM_RUNTIME_FLAG_IS_SET(vm, flag) ((vm)->runtimeFlags & (flag))
@@ -426,9 +554,11 @@ enum {
 
 #define SCM_CCONT_DATA_SIZE 6
 
+typedef ScmObj ScmCContinuationProc(ScmObj result, void **data);
+
 typedef struct ScmCContinuation {
     SCM_HEADER;
-    ScmObj (*func)(ScmObj value, void **data);
+    ScmCContinuationProc *func;
     void *data[SCM_CCONT_DATA_SIZE];
 } ScmCContinuation;
 
@@ -438,33 +568,41 @@ typedef struct ScmCContinuation {
 SCM_CLASS_DECL(Scm_CContClass);
 #define SCM_CLASS_CCONT           (&Scm_CContClass)
 
-SCM_EXTERN void Scm_VMPushCC(ScmObj (*func)(ScmObj value, void **data),
+SCM_EXTERN void Scm_VMPushCC(ScmCContinuationProc *func,
 			     void **data,
 			     int datasize);
 
-/*
- * Compiler context
- */
-
-enum {
-    SCM_COMPILE_STMT,           /* Statement context.  The value of this
-                                   expression will be discarded. */
-    SCM_COMPILE_TAIL,           /* This is a tail expression. */
-    SCM_COMPILE_NORMAL          /* Normal calling sequence. */
-};
-
 /*
  * Compiler flags
  */
 
 enum {
-    SCM_COMPILE_NOINLINE = (1L<<0), /* Do not inline procedures */
-    SCM_COMPILE_NOSOURCE = (1L<<1), /* Do not insert source info */
-    SCM_COMPILE_SHOWRESULT = (1L<<2), /* Display each result of compilation */
+    SCM_COMPILE_NOINLINE_GLOBALS = (1L<<0),/* Do not inline global procs */
+    SCM_COMPILE_NOINLINE_LOCALS = (1L<<1), /* Do not inline local procs */
+    SCM_COMPILE_NOINLINE_CONSTS = (1L<<2), /* Do not inline constants */
+    SCM_COMPILE_NOSOURCE = (1L<<3),        /* Do not insert source info */
+    SCM_COMPILE_SHOWRESULT = (1L<<4),      /* Display each result of
+                                              compilation */
+    SCM_COMPILE_NOCOMBINE = (1L<<5)        /* Do not combine instructions */
 };
 
 #define SCM_VM_COMPILER_FLAG_IS_SET(vm, flag) ((vm)->compilerFlags & (flag))
 #define SCM_VM_COMPILER_FLAG_SET(vm, flag)    ((vm)->compilerFlags |= (flag))
 #define SCM_VM_COMPILER_FLAG_CLEAR(vm, flag)  ((vm)->compilerFlags &= ~(flag))
 
+/*
+ * Compiler internal APIs
+ */
+
+SCM_EXTERN ScmObj Scm_Compile(ScmObj program, ScmObj mod);
+
+SCM_EXTERN ScmObj Scm_CallSyntaxCompiler(ScmObj syn, ScmObj from, ScmObj env);
+SCM_EXTERN ScmObj Scm_CallMacroExpander(ScmMacro *mac, ScmObj expr, ScmObj env);
+SCM_EXTERN ScmObj Scm_CallMacroExpanderOld(ScmMacro *mac, ScmObj expr, ScmObj env);
+SCM_EXTERN int    Scm_HasInlinerP(ScmObj obj);
+SCM_EXTERN ScmObj Scm_CallProcedureInliner(ScmObj obj, ScmObj form, ScmObj env);
+
+/* This is in module.c, but it's not for public, so declaration is here. */
+SCM_EXTERN ScmModule* Scm_GaucheInternalModule(void);
+
 #endif /* GAUCHE_VM_H */
