@@ -1,7 +1,7 @@
 /*
  * load.c - load a program
  *
- *   Copyright (c) 2000-2005 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2007  Shiro Kawai  <shiro@acm.org>
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: load.c,v 1.106 2006/04/06 21:41:25 shirok Exp $
+ *  $Id: load.c,v 1.113 2007/04/12 03:26:55 shirok Exp $
  */
 
 #include <stdlib.h>
@@ -58,7 +58,6 @@ static struct {
     ScmGloc *load_path_rec;     /* *load-path*         */
     ScmGloc *dynload_path_rec;  /* *dynamic-load-path* */
     ScmGloc *load_suffixes_rec; /* *load-suffixes*     */
-    ScmGloc *cond_features_rec; /* *cond-features*     */
     ScmInternalMutex path_mutex;
 
     /* Provided features */
@@ -84,6 +83,15 @@ static ScmObj key_environment        = SCM_UNBOUND;
 static ScmObj key_macro              = SCM_UNBOUND;
 static ScmObj key_ignore_coding      = SCM_UNBOUND;
 
+/* small utility */
+static void load_packet_init(ScmLoadPacket *packet)
+{
+    if (packet) {
+        packet->exception = SCM_FALSE;
+        packet->loaded = FALSE;
+    }
+}
+
 /*--------------------------------------------------------------------
  * Scm_LoadFromPort
  * 
@@ -172,6 +180,8 @@ ScmObj Scm_VMLoadFromPort(ScmPort *port, ScmObj next_paths,
     /* Sanity check */
     if (!SCM_IPORTP(port))
         Scm_Error("input port required, but got: %S", port);
+
+
     if (SCM_PORT_CLOSED_P(port))
         Scm_Error("port already closed: %S", port);
     if (SCM_MODULEP(env)) {
@@ -231,9 +241,25 @@ static SCM_DEFINE_SUBR(load_from_port_STUB, 1, 1,
                        SCM_OBJ(&load_from_port_NAME), load_from_port,
                        NULL, NULL);
 
-void Scm_LoadFromPort(ScmPort *port, int flags)
+int Scm_LoadFromPort(ScmPort *port, int flags, ScmLoadPacket *packet)
 {
-    Scm_Apply(SCM_OBJ(&load_from_port_STUB), SCM_LIST1(SCM_OBJ(port)));
+    ScmEvalPacket eresult;
+    int r;
+
+    load_packet_init(packet);
+    if (flags&SCM_LOAD_PROPAGATE_ERROR) {
+        Scm_ApplyRec(SCM_OBJ(&load_from_port_STUB), SCM_LIST1(SCM_OBJ(port)));
+        if (packet) packet->loaded = TRUE;
+        return 0;
+    } else {
+        r = Scm_Apply(SCM_OBJ(&load_from_port_STUB), SCM_LIST1(SCM_OBJ(port)),
+                      &eresult);
+        if (packet) {
+            packet->exception = eresult.exception;
+            packet->loaded = (r >= 0);
+        }
+        return (r < 0)? -1 : 0;
+    }
 }
 
 /*---------------------------------------------------------------------
@@ -393,7 +419,7 @@ static ScmObj load(ScmObj *args, int argc, void *data)
     ScmString *file;
     ScmObj paths, env;
     int flags = 0;
-    
+
     if (!SCM_STRINGP(args[0])) {
         Scm_Error("string required, but got %S", args[0]);
     }
@@ -411,10 +437,11 @@ static SCM_DEFINE_STRING_CONST(load_NAME, "load", 4, 4);
 static SCM_DEFINE_SUBR(load_STUB, 1, 1, SCM_OBJ(&load_NAME), load, NULL, NULL);
 
 
-int Scm_Load(const char *cpath, int flags)
+int Scm_Load(const char *cpath, int flags, ScmLoadPacket *packet)
 {
-    ScmObj r, f = SCM_MAKE_STR_COPYING(cpath);
+    ScmObj f = SCM_MAKE_STR_COPYING(cpath);
     ScmObj options = SCM_NIL;
+    ScmEvalPacket eresult;
     
     if (flags&SCM_LOAD_QUIET_NOFILE) {
         options = Scm_Cons(key_error_if_not_found,
@@ -424,9 +451,22 @@ int Scm_Load(const char *cpath, int flags)
         options = Scm_Cons(key_ignore_coding,
                            Scm_Cons(SCM_TRUE, options));
     }
-    
-    r = Scm_Apply(SCM_OBJ(&load_STUB), Scm_Cons(f, options));
-    return !SCM_FALSEP(r);
+
+    load_packet_init(packet);
+    if (flags&SCM_LOAD_PROPAGATE_ERROR) {
+        ScmObj r = Scm_ApplyRec(SCM_OBJ(&load_STUB), Scm_Cons(f, options));
+        if (packet) {
+            packet->loaded = !SCM_FALSEP(r);
+        }
+        return 0;
+    } else {
+        int r = Scm_Apply(SCM_OBJ(&load_STUB), Scm_Cons(f, options), &eresult);
+        if (packet) {
+            packet->exception = eresult.exception;
+            packet->loaded = (r > 0 && !SCM_FALSEP(eresult.results[0]));
+        }
+        return (r >= 0)? 0 : -1;
+    }
 }
 
 /*
@@ -525,59 +565,6 @@ ScmObj Scm_AddLoadPath(const char *cpath, int afterp)
     return r;
 }
 
-#if 0 /* not sure we need this yet */
-/*------------------------------------------------------------------
- * Append FEATURE to *cond-features*.  Once added, the symbol FEATURE
- * will be recognized by cond-expand.
- *
- *  (cond-expand (FEATURE body ...)) 
- *   
- * If loading a module is required in order to make FEATURE available,
- * such module can be specified in MODULE argument.  It can be SCM_FALSE
- * if the feature is built-in.
- *
- * Although cond-expand can be used like #ifdefs of cpp, it is a bit
- * different since you can test a feature _before_ loading the module
- * (or, you can say testing the feature itself implies loading the
- * module if available).  Thus it is not recommended to add a feature
- * as a result of loading a specific module.  One possible usage is
- * to add a feature conditionally if a module may or may not provide
- * such a feature depending on platforms or compile-time options.
- * (An example of this is gauche.net, which may provide gauche-ipv6
- * if IPv6 support is compiled in.)
- *
- * Because of this obscurity, I do not encourage user-land programs
- * to use this API.  We may try some use within Gauche bundled extensions,
- * and see how well it works, before we make it open to public.
- * (Esp. we need to come up a reasonable naming scheme of features
- * if we allow arbitrary extensions to add them).
- */
-
-ScmObj
-Scm_AddCondFeature(ScmObj feature, ScmObj module)
-{
-    ScmObj cell, z;
-   
-    if (!SCM_SYMBOLP(feature)) {
-        Scm_Error("feature name must be a symbol, but got %S", feature);
-    }
-    if (!SCM_SYMBOLP(module) && !SCM_FALSEP(module)) {
-        Scm_Error("module must be a symbol or #f, but got %S", module);
-    }
-    if (SCM_FALSEP(module)) {
-        cell = SCM_LIST1(feature);
-    } else {
-        cell = SCM_LIST2(feature, module);
-    }
-    
-    (void)SCM_INTERNAL_MUTEX_LOCK(ldinfo.path_mutex);
-    z = ldinfo.cond_features_rec->value =
-        Scm_Cons(cell, ldinfo.cond_features_rec->value);
-    (void)SCM_INTERNAL_MUTEX_UNLOCK(ldinfo.path_mutex);
-    return z;
-}
-#endif
-
 /*------------------------------------------------------------------
  * Dynamic link
  */
@@ -873,15 +860,22 @@ ScmObj Scm_DynLoad(ScmString *filename, ScmObj initfn, int export_)
  *   the calling thread.
  */
 
-ScmObj Scm_Require(ScmObj feature)
+int Scm_Require(ScmObj feature, int flags, ScmLoadPacket *packet)
 {
     ScmObj filename;
     ScmVM *vm = Scm_VM();
     ScmObj provided, providing, p, q;
-    int loop = FALSE;
+    int loop = FALSE, r;
+    ScmLoadPacket xresult;
 
+    load_packet_init(packet);
     if (!SCM_STRINGP(feature)) {
-        Scm_Error("require: string expected, but got %S\n", feature);
+        ScmObj e = Scm_MakeError(Scm_Sprintf("require: string expected, but got %S\n", feature));
+        if (flags&SCM_LOAD_PROPAGATE_ERROR) Scm_Raise(e);
+        else {
+            if (packet) packet->exception = e;
+            return -1;
+        }
     }
 
     (void)SCM_INTERNAL_MUTEX_LOCK(ldinfo.prov_mutex);
@@ -921,23 +915,35 @@ ScmObj Scm_Require(ScmObj feature)
     }
     (void)SCM_INTERNAL_MUTEX_UNLOCK(ldinfo.prov_mutex);
 
-    if (loop) Scm_Error("a loop is detected in the require dependency involving feature %S", feature);
-    if (!SCM_FALSEP(provided)) return SCM_TRUE;
-    SCM_UNWIND_PROTECT {
-        filename = Scm_StringAppendC(SCM_STRING(feature), ".scm", 4, 4);
-        Scm_Load(Scm_GetStringConst(SCM_STRING(filename)), 0);
-    } SCM_WHEN_ERROR {
+    if (loop) {
+        ScmObj e = Scm_MakeError(Scm_Sprintf("a loop is detected in the require dependency involving feature %S", feature));
+        if (flags&SCM_LOAD_PROPAGATE_ERROR) Scm_Raise(e);
+        else {
+            if (packet) packet->exception = e;
+            return -1;
+        }
+    }
+        
+    if (!SCM_FALSEP(provided)) return 0;
+    filename = Scm_StringAppendC(SCM_STRING(feature), ".scm", 4, 4);
+    r = Scm_Load(Scm_GetStringConst(SCM_STRING(filename)), 0, &xresult);
+    if (packet) packet->exception = xresult.exception;
+
+    if (r < 0) {
         (void)SCM_INTERNAL_MUTEX_LOCK(ldinfo.prov_mutex);
         ldinfo.providing = Scm_AssocDeleteX(feature, ldinfo.providing, SCM_CMP_EQUAL);
         (void)SCM_INTERNAL_COND_SIGNAL(ldinfo.prov_cv);
         (void)SCM_INTERNAL_MUTEX_UNLOCK(ldinfo.prov_mutex);
-        SCM_NEXT_HANDLER;
-    } SCM_END_PROTECT;
+        if (flags&SCM_LOAD_PROPAGATE_ERROR) Scm_Raise(xresult.exception);
+        else return -1;
+    }
+
     (void)SCM_INTERNAL_MUTEX_LOCK(ldinfo.prov_mutex);
     ldinfo.providing = Scm_AssocDeleteX(feature, ldinfo.providing, SCM_CMP_EQUAL);
     (void)SCM_INTERNAL_COND_SIGNAL(ldinfo.prov_cv);
     (void)SCM_INTERNAL_MUTEX_UNLOCK(ldinfo.prov_mutex);
-    return SCM_TRUE;
+    if (packet) packet->loaded = TRUE;
+    return 0;
 }
 
 ScmObj Scm_Provide(ScmObj feature)
@@ -1080,7 +1086,7 @@ ScmObj Scm_LoadAutoload(ScmAutoload *adata)
     prev_module = vm->module;
     SCM_UNWIND_PROTECT {
         vm->module = adata->module;
-        Scm_Require(SCM_OBJ(adata->path));
+        Scm_Require(SCM_OBJ(adata->path), SCM_LOAD_PROPAGATE_ERROR, NULL);
         vm->module = prev_module;
     
         if (adata->import_from) {
@@ -1127,6 +1133,27 @@ ScmObj Scm_LoadAutoload(ScmAutoload *adata)
     return adata->value;
 }
 
+/*------------------------------------------------------------------
+ * Compatibility stuff
+ */
+
+void Scm__LoadFromPortCompat(ScmPort *port, int flags)
+{
+    Scm_LoadFromPort(port, flags|SCM_LOAD_PROPAGATE_ERROR, NULL);
+}
+
+int  Scm__LoadCompat(const char *file, int flags)
+{
+    return (0 == Scm_Load(file, flags|SCM_LOAD_PROPAGATE_ERROR, NULL));
+}
+
+ScmObj Scm__RequireCompat(ScmObj feature)
+{
+    Scm_Require(feature, SCM_LOAD_PROPAGATE_ERROR, NULL);
+    return SCM_TRUE;
+}
+
+
 /*------------------------------------------------------------------
  * Initialization
  */
@@ -1134,8 +1161,7 @@ ScmObj Scm_LoadAutoload(ScmAutoload *adata)
 void Scm__InitLoad(void)
 {
     ScmModule *m = Scm_SchemeModule();
-    ScmObj init_load_path, init_dynload_path, init_load_suffixes,
-        init_cond_features, t;
+    ScmObj init_load_path, init_dynload_path, init_load_suffixes, t;
 
     init_load_path = t = SCM_NIL;
     SCM_APPEND(init_load_path, t, break_env_paths("GAUCHE_LOAD_PATH"));
@@ -1150,21 +1176,6 @@ void Scm__InitLoad(void)
     init_load_suffixes = t = SCM_NIL;
     SCM_APPEND1(init_load_suffixes, t, SCM_MAKE_STR(LOAD_SUFFIX));
 
-    init_cond_features = t = SCM_NIL;
-    SCM_APPEND1(init_cond_features, t, SCM_LIST1(SCM_SYM_GAUCHE));
-#ifdef __MINGW32__
-    SCM_APPEND1(init_cond_features, t, SCM_LIST1(SCM_SYM_GAUCHE_WINDOWS));
-#endif /*__MINGW32__*/
-#if defined(GAUCHE_CHAR_ENCODING_EUC_JP)
-    SCM_APPEND1(init_cond_features, t, SCM_LIST1(SCM_SYM_GAUCHE_EUCJP));
-#elif defined(GAUCHE_CHAR_ENCODING_SJIS)
-    SCM_APPEND1(init_cond_features, t, SCM_LIST1(SCM_SYM_GAUCHE_SJIS));
-#elif defined(GAUCHE_CHAR_ENCODING_UTF8)
-    SCM_APPEND1(init_cond_features, t, SCM_LIST1(SCM_SYM_GAUCHE_UTF8));
-#else
-    SCM_APPEND1(init_cond_features, t, SCM_LIST1(SCM_SYM_GAUCHE_NONE));
-#endif
-
     (void)SCM_INTERNAL_MUTEX_INIT(ldinfo.path_mutex);
     (void)SCM_INTERNAL_MUTEX_INIT(ldinfo.prov_mutex);
     (void)SCM_INTERNAL_COND_INIT(ldinfo.prov_cv);
@@ -1185,7 +1196,6 @@ void Scm__InitLoad(void)
     DEF(ldinfo.load_path_rec,    SCM_SYM_LOAD_PATH, init_load_path);
     DEF(ldinfo.dynload_path_rec, SCM_SYM_DYNAMIC_LOAD_PATH, init_dynload_path);
     DEF(ldinfo.load_suffixes_rec, SCM_SYM_LOAD_SUFFIXES, init_load_suffixes);
-    DEF(ldinfo.cond_features_rec, SCM_SYM_COND_FEATURES, init_cond_features);
 
     ldinfo.provided =
         SCM_LIST5(SCM_MAKE_STR("srfi-2"), /* and-let* */
