@@ -1,7 +1,7 @@
 /*
  * macro.c - macro implementation
  *
- *   Copyright (c) 2000-2005 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2007  Shiro Kawai  <shiro@acm.org>
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: macro.c,v 1.60 2006/03/27 09:17:25 shirok Exp $
+ *  $Id: macro.c,v 1.66 2007/03/02 07:39:13 shirok Exp $
  */
 
 #define LIBGAUCHE_BODY
@@ -41,7 +41,7 @@
 #include "gauche/builtin-syms.h"
 
 /* avoid C++ reserved name conflict.
-   (I hate languages that takes away names from programmers!) */
+   (I hate languages that take away names from programmers!) */
 #define template templat
 
 /* define if you want to debug syntax-rule expander */
@@ -188,7 +188,7 @@ static ScmObj macro_transform(ScmObj self, ScmObj form, ScmObj env,
 {
     ScmObj proc = SCM_OBJ(data);
     SCM_ASSERT(SCM_SYNTACTIC_CLOSURE_P(form));
-    return Scm_Apply(proc, SCM_LIST1(form));
+    return Scm_ApplyRec(proc, SCM_LIST1(form));
 }
 
 ScmObj Scm_MakeMacroTransformer(ScmSymbol *name, ScmObj proc)
