@@ -1,7 +1,7 @@
 /*
  * portapi.c - port common API
  *
- *   Copyright (c) 2000-2003 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2007  Shiro Kawai  <shiro@acm.org>
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,17 +30,38 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: portapi.c,v 1.19 2004/05/21 05:02:12 shirok Exp $
+ *  $Id: portapi.c,v 1.31 2007/03/02 07:39:14 shirok Exp $
  */
 
-/* This file is included twice by port.c to define safe- and unsafe-
- * variant of port common APIs.
+/* This file is included _twice_ by port.c to define safe- and unsafe-
+ * variant of port common APIs.  It is to minimize the overhead of
+ * locking operations.
  *
  * The macro SHORTCUT allows 'safe' version to bypass lock/unlock
  * stuff by calling 'unsafe' version when the port is already locked by
  * the calling thread.
  */
 
+/* [scratch and ungottern buffer]
+ *   It is always possible to mix binary and character i/o for Gauche's
+ *   ports.  To support peek operations in Scheme and 'unget' operations
+ *   in C, we need to buffer at most one character or its equivalent
+ *   byte sequence.   The 'ungotten' and 'scratch' fields are used for
+ *   character and binary buffering, respectively.
+ *   (This level of buffering is common to all input port types, and
+ *   distinct from the buffering of 'buffered' (file) port type.)
+ *
+ *   'Ungotten' field keeps SCM_CHAR_INVALID if there's no buffered
+ *   character.  Otherwise, its value is the buffered character.
+ *   The number of bytes in the 'scratch' array is kept in 'scrcnt'
+ *   field.  If 'scrcnt' field is not zero, there's data in the
+ *   'scratch' array.
+ *
+ *   In no cases there should be data in both ungotten and scratch
+ *   field.  The consistency is taken care of the routines defined here;
+ *   no other routine should touch these buffering field.
+ */
+
 #ifdef SAFE_PORT_OP
 #define VMDECL        ScmVM *vm = Scm_VM()
 #define LOCK(p)       PORT_LOCK(p, vm)
@@ -56,15 +77,18 @@
 #define SHORTCUT(p, unsafe) /* none */
 #endif
 
-/* Convenience macro */
 
+/* Convenience macro */
+#ifndef CLOSE_CHECK
 #define CLOSE_CHECK(port)                                               \
     do {                                                                \
         if (SCM_PORT_CLOSED_P(port)) {                                  \
             UNLOCK(p);                                                  \
-            Scm_Error("I/O attempted on closed port: %S", (port));      \
+            Scm_PortError((port), SCM_PORT_ERROR_CLOSED,                \
+                          "I/O attempted on closed port: %S", (port));  \
         }                                                               \
     } while (0)
+#endif /* CLOSE_CHECK */
 
 /*=================================================================
  * Putb
@@ -84,7 +108,7 @@ void Scm_PutbUnsafe(ScmByte b, ScmPort *p)
     switch (SCM_PORT_TYPE(p)) {
     case SCM_PORT_FILE:
         if (p->src.buf.current >= p->src.buf.end) {
-            SAFE_CALL(p, bufport_flush(p, 1, FALSE));
+            SAFE_CALL(p, bufport_flush(p, p->src.buf.current - p->src.buf.buffer, FALSE));
         }
         SCM_ASSERT(p->src.buf.current < p->src.buf.end);
         *p->src.buf.current++ = b;
@@ -103,7 +127,8 @@ void Scm_PutbUnsafe(ScmByte b, ScmPort *p)
         break;
     default:
         UNLOCK(p);
-        Scm_Error("bad port type for output: %S", p);
+        Scm_PortError(p, SCM_PORT_ERROR_OUTPUT,
+                      "bad port type for output: %S", p);
     }
 }
 
@@ -127,7 +152,7 @@ void Scm_PutcUnsafe(ScmChar c, ScmPort *p)
     case SCM_PORT_FILE:
         nb = SCM_CHAR_NBYTES(c);
         if (p->src.buf.current+nb > p->src.buf.end) {
-            SAFE_CALL(p, bufport_flush(p, nb, FALSE));
+            SAFE_CALL(p, bufport_flush(p, p->src.buf.current - p->src.buf.buffer, FALSE));
         }
         SCM_ASSERT(p->src.buf.current+nb <= p->src.buf.end);
         SCM_CHAR_PUT(p->src.buf.current, c);
@@ -146,12 +171,13 @@ void Scm_PutcUnsafe(ScmChar c, ScmPort *p)
         UNLOCK(p);
         break;
     case SCM_PORT_PROC:
-        PORT_SAFE_CALL(p, p->src.vt.Putc(c, p));
+        SAFE_CALL(p, p->src.vt.Putc(c, p));
         UNLOCK(p);
         break;
     default:
         UNLOCK(p);
-        Scm_Error("bad port type for output: %S", p);
+        Scm_PortError(p, SCM_PORT_ERROR_OUTPUT,
+                      "bad port type for output: %S", p);
     }
 }
 
@@ -171,8 +197,10 @@ void Scm_PutsUnsafe(ScmString *s, ScmPort *p)
     CLOSE_CHECK(p);
     
     switch (SCM_PORT_TYPE(p)) {
-    case SCM_PORT_FILE:
-        SAFE_CALL(p, bufport_write(p, SCM_STRING_START(s), SCM_STRING_SIZE(s)));
+    case SCM_PORT_FILE: {
+        u_int size;
+        const char *ss = Scm_GetStringContent(s, &size, NULL, NULL);
+        SAFE_CALL(p, bufport_write(p, ss, size));
         
         if (p->src.buf.mode == SCM_PORT_BUFFER_LINE) {
             const char *cp = p->src.buf.current;
@@ -187,6 +215,7 @@ void Scm_PutsUnsafe(ScmString *s, ScmPort *p)
         }
         UNLOCK(p);
         break;
+    }
     case SCM_PORT_OSTR:
         Scm_DStringAdd(&p->src.ostr, s);
         UNLOCK(p);
@@ -197,7 +226,8 @@ void Scm_PutsUnsafe(ScmString *s, ScmPort *p)
         break;
     default:
         UNLOCK(p);
-        Scm_Error("bad port type for output: %S", p);
+        Scm_PortError(p, SCM_PORT_ERROR_OUTPUT,
+                      "bad port type for output: %S", p);
     }
 }
 
@@ -242,7 +272,8 @@ void Scm_PutzUnsafe(const char *s, int siz, ScmPort *p)
         break;
     default:
         UNLOCK(p);
-        Scm_Error("bad port type for output: %S", p);
+        Scm_PortError(p, SCM_PORT_ERROR_OUTPUT,
+                      "bad port type for output: %S", p);
     }
 }
 
@@ -274,7 +305,8 @@ void Scm_FlushUnsafe(ScmPort *p)
         break;
     default:
         UNLOCK(p);
-        Scm_Error("bad port type for output: %S", p);
+        Scm_PortError(p, SCM_PORT_ERROR_OUTPUT,
+                      "bad port type for output: %S", p);
     }
 }
 
@@ -291,7 +323,12 @@ void Scm_UngetcUnsafe(ScmChar c, ScmPort *p)
     VMDECL;
     SHORTCUT(p, Scm_UngetcUnsafe(c, p); return);
     LOCK(p);
-    SCM_UNGETC(c, p);
+    if (p->ungotten != SCM_CHAR_INVALID
+        || p->scrcnt != 0) {
+        Scm_PortError(p, SCM_PORT_ERROR_INPUT,
+                      "pushback buffer overflow on port %S", p);
+    }
+    p->ungotten = c;
     UNLOCK(p);
 }
 
@@ -305,9 +342,9 @@ ScmChar Scm_PeekcUnsafe(ScmPort *p)
     VMDECL;
     SHORTCUT(p, return Scm_PeekcUnsafe(p));
     LOCK(p);
-    if ((ch = SCM_PORT_UNGOTTEN(p)) == SCM_CHAR_INVALID) {
+    if ((ch = p->ungotten) == SCM_CHAR_INVALID) {
         ch = Scm_GetcUnsafe(p);
-        SCM_PORT_UNGOTTEN(p) = ch;
+        p->ungotten = ch;
     }
     UNLOCK(p);
     return ch;
@@ -326,9 +363,12 @@ void Scm_UngetbUnsafe(int b, ScmPort *p)
     VMDECL;
     SHORTCUT(p, Scm_UngetbUnsafe(b, p); return);
     LOCK(p);
-    /* caller must ensure ungetb is called only at the fresh state */
-    p->scratch[0] = b;
-    p->scrcnt = 1;
+    if (p->ungotten != SCM_CHAR_INVALID
+        || p->scrcnt >= SCM_CHAR_MAX_BYTES) {
+        Scm_PortError(p, SCM_PORT_ERROR_INPUT,
+                      "pushback buffer overflow on port %S", p);
+    }
+    p->scratch[p->scrcnt++] = b;
     UNLOCK(p);
 }
 
@@ -371,16 +411,23 @@ int Scm_PeekbUnsafe(ScmPort *p)
  */
 
 #ifndef SHIFT_SCRATCH  /* we need to define this only once */
+#define SHIFT_SCRATCH
+
 /* shift scratch buffer content */
-#define SHIFT_SCRATCH(p, off) \
-   do { int i_; for (i_=0; i_ < (p)->scrcnt; i_++) (p)->scratch[i_]=(p)->scratch[i_+(off)]; } while (0)
+static inline void shift_scratch(ScmPort *p, int off)
+{
+    int i;
+    for (i=0; i<p->scrcnt; i++) {
+        p->scratch[i] = p->scratch[i+off];
+    }
+}
 
 /* handle the case that there's remaining data in the scratch buffer */
 static int getb_scratch(ScmPort *p)
 {
     int b = (unsigned char)p->scratch[0];
     p->scrcnt--;
-    SHIFT_SCRATCH(p, 1);
+    shift_scratch(p, 1);
     return b;
 }
 
@@ -433,7 +480,8 @@ int Scm_GetbUnsafe(ScmPort *p)
             break;
         default:
             UNLOCK(p);
-            Scm_Error("bad port type for output: %S", p);
+            Scm_PortError(p, SCM_PORT_ERROR_INPUT,
+                          "bad port type for input: %S", p);
         }
     }
     UNLOCK(p);
@@ -463,7 +511,8 @@ static int getc_scratch_unsafe(ScmPort *p)
         SAFE_CALL(p, r = Scm_Getb(p));
         if (r == EOF) {
             UNLOCK(p);
-            Scm_Error("encountered EOF in middle of a multibyte character from port %S", p);
+            Scm_PortError(p, SCM_PORT_ERROR_INPUT,
+                          "encountered EOF in middle of a multibyte character from port %S", p);
         }
         tbuf[i] = (char)r;
     }
@@ -521,7 +570,8 @@ int Scm_GetcUnsafe(ScmPort *p)
                     if (filled <= 0) {
                         /* TODO: make this behavior customizable */
                         UNLOCK(p);
-                        Scm_Error("encountered EOF in middle of a multibyte character from port %S", p);
+                        Scm_PortError(p, SCM_PORT_ERROR_INPUT,
+                                      "encountered EOF in middle of a multibyte character from port %S", p);
                     }
                     if (filled >= rest) {
                         memcpy(p->scratch+p->scrcnt, p->src.buf.current, rest);
@@ -543,7 +593,7 @@ int Scm_GetcUnsafe(ScmPort *p)
             }
         } else {
             c = first;
-            if (c == '\n') p->src.buf.line++;
+            if (c == '\n') p->line++;
         }
         UNLOCK(p);
         return c;
@@ -558,22 +608,26 @@ int Scm_GetcUnsafe(ScmPort *p)
             if (p->src.istr.current + nb > p->src.istr.end) {
                 /* TODO: make this behavior customizable */
                 UNLOCK(p);
-                Scm_Error("encountered EOF in middle of a multibyte character from port %S", p);
+                Scm_PortError(p, SCM_PORT_ERROR_INPUT,
+                              "encountered EOF in middle of a multibyte character from port %S", p);
             }
             SCM_CHAR_GET(p->src.istr.current-1, c);
             p->src.istr.current += nb;
         } else {
             c = first;
+            if (c == '\n') p->line++;
         }
         UNLOCK(p);
         return c;
     case SCM_PORT_PROC:
         SAFE_CALL(p, c = p->src.vt.Getc(p));
+        if (c == '\n') p->line++;
         UNLOCK(p);
         return c;
     default:
         UNLOCK(p);
-        Scm_Error("bad port type for output: %S", p);
+        Scm_PortError(p, SCM_PORT_ERROR_INPUT,
+                      "bad port type for input: %S", p);
     }
     return 0;/*dummy*/
 }
@@ -599,7 +653,7 @@ static int getz_scratch_unsafe(char *buf, int buflen, ScmPort *p)
     if (p->scrcnt >= buflen) {
         memcpy(buf, p->scratch, buflen);
         p->scrcnt -= buflen;
-        SHIFT_SCRATCH(p, buflen);
+        shift_scratch(p, buflen);
         return buflen;
     } else {
         memcpy(buf, p->scratch, p->scrcnt);
@@ -671,7 +725,8 @@ int Scm_GetzUnsafe(char *buf, int buflen, ScmPort *p)
         return r;
     default:
         UNLOCK(p);
-        Scm_Error("bad port type for output: %S", p);
+        Scm_PortError(p, SCM_PORT_ERROR_INPUT,
+                      "bad port type for input: %S", p);
     }
     return -1;                  /* dummy */
 }
@@ -708,7 +763,7 @@ ScmObj readline_body(ScmPort *p)
     b1 = Scm_GetbUnsafe(p);
     if (b1 == EOF) return SCM_EOF;
     for (;;) {
-        if (b1 == EOF) return Scm_DStringGet(&ds);
+        if (b1 == EOF) return Scm_DStringGet(&ds, 0);
         if (b1 == '\n') break;
         if (b1 == '\r') {
             b2 = Scm_GetbUnsafe(p);
@@ -719,8 +774,8 @@ ScmObj readline_body(ScmPort *p)
         SCM_DSTRING_PUTB(&ds, b1);
         b1 = Scm_GetbUnsafe(p);
     }
-    if(SCM_PORT_TYPE(p) == SCM_PORT_FILE) p->src.buf.line++;
-    return Scm_DStringGet(&ds);
+    p->line++;
+    return Scm_DStringGet(&ds, 0);
 }
 #endif /* READLINE_AUX */
 
@@ -740,6 +795,44 @@ ScmObj Scm_ReadLineUnsafe(ScmPort *p)
     return r;
 }
 
+/*=================================================================
+ * ByteReady
+ */
+
+#ifdef SAFE_PORT_OP
+int Scm_ByteReady(ScmPort *p)
+#else
+int Scm_ByteReadyUnsafe(ScmPort *p)
+#endif
+{
+    int r = 0;
+    VMDECL;
+    SHORTCUT(p, return Scm_ByteReadyUnsafe(p));
+    if (!SCM_IPORTP(p)) Scm_Error("input port required, but got %S", p);
+    LOCK(p);
+    if (p->ungotten != SCM_CHAR_INVALID
+        || p->scrcnt > 0) {
+        r = TRUE;
+    } else {
+        switch (SCM_PORT_TYPE(p)) {
+        case SCM_PORT_FILE:
+            if (p->src.buf.current < p->src.buf.end) r = TRUE;
+            else if (p->src.buf.ready == NULL) r = TRUE;
+            else {
+                SAFE_CALL(p, r = (p->src.buf.ready(p) != SCM_FD_WOULDBLOCK));
+            }
+            break;
+        case SCM_PORT_PROC:
+            SAFE_CALL(p, r = p->src.vt.Ready(p, FALSE));
+            break;
+        default:
+            r = TRUE;
+        }
+    }
+    UNLOCK(p);
+    return r;
+}
+
 /*=================================================================
  * CharReady
  */
@@ -755,7 +848,7 @@ int Scm_CharReadyUnsafe(ScmPort *p)
     SHORTCUT(p, return Scm_CharReadyUnsafe(p));
     if (!SCM_IPORTP(p)) Scm_Error("input port required, but got %S", p);
     LOCK(p);
-    if (SCM_PORT_UNGOTTEN(p) != SCM_CHAR_INVALID) r = TRUE;
+    if (p->ungotten != SCM_CHAR_INVALID) r = TRUE;
     else {
         switch (SCM_PORT_TYPE(p)) {
         case SCM_PORT_FILE:
@@ -766,7 +859,7 @@ int Scm_CharReadyUnsafe(ScmPort *p)
             }
             break;
         case SCM_PORT_PROC:
-            SAFE_CALL(p, r = p->src.vt.Ready(p));
+            SAFE_CALL(p, r = p->src.vt.Ready(p, TRUE));
             break;
         default:
             r = TRUE;
@@ -817,7 +910,8 @@ ScmObj Scm_PortSeekUnsafe(ScmPort *p, ScmObj off, int whence)
     VMDECL;
     SHORTCUT(p, return Scm_PortSeekUnsafe(p, off, whence));
     if (SCM_PORT_CLOSED_P(p)) {
-        Scm_Error("attempt to seek on closed port: %S", p);
+        Scm_PortError(p, SCM_PORT_ERROR_CLOSED,
+                      "attempt to seek on closed port: %S", p);
     }
     LOCK(p);
     switch (SCM_PORT_TYPE(p)) {
