@@ -1,7 +1,7 @@
 /*
  * read.c - reader
  *
- *   Copyright (c) 2000-2005 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2007  Shiro Kawai  <shiro@acm.org>
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: read.c,v 1.83 2005/10/13 08:14:13 shirok Exp $
+ *  $Id: read.c,v 1.92 2007/04/12 03:26:55 shirok Exp $
  */
 
 #include <stdio.h>
@@ -203,7 +203,8 @@ void Scm_ReadError(ScmPort *port, const char *msg, ...)
     Scm_Vprintf(SCM_PORT(ostr), msg, ap, TRUE);
     va_end(ap);
 
-    rerr = Scm_MakeReadError(Scm_GetOutputString(SCM_PORT(ostr)), port, line);
+    rerr = Scm_MakeReadError(Scm_GetOutputString(SCM_PORT(ostr), 0),
+                             port, line);
     Scm_Raise(rerr);
 }
 
@@ -238,7 +239,7 @@ static void ref_push(ScmReadContext *ctx, ScmObj obj, ScmObj finisher)
     ctx->pending = Scm_Acons(obj, finisher, ctx->pending);
 }
 
-static ScmObj ref_val(ref)
+static ScmObj ref_val(ScmObj ref)
 {
     if (!SCM_READ_REFERENCE_REALIZED(ref)) {
         Scm_Error("reader encontered unresolved read reference.  Implementation error?");
@@ -270,7 +271,7 @@ static void read_context_flush(ScmReadContext *ctx)
         finisher = SCM_CDR(entry);
 
         if (!SCM_FALSEP(finisher)) {
-            Scm_Apply(finisher, SCM_LIST1(obj));
+            Scm_ApplyRec(finisher, SCM_LIST1(obj));
         } else if (SCM_PAIRP(obj)) {
             SCM_FOR_EACH(ep, obj) {
                 if (SCM_READ_REFERENCE_P(SCM_CAR(ep))) {
@@ -484,8 +485,13 @@ static ScmObj read_internal(ScmPort *port, ScmReadContext *ctx)
                 }
             case ';':
                 /* #;expr - comment out sexpr */
-                read_item(port, ctx); /* read and discard */
-                return SCM_UNDEFINED; /* indicate this is a comment */
+                {
+                    int orig = ctx->flags;
+                    ctx->flags |= SCM_READ_DISABLE_CTOR;
+                    read_item(port, ctx); /* read and discard */
+                    ctx->flags = orig;
+                    return SCM_UNDEFINED; /* indicate this is a comment */
+                }
             default:
                 Scm_ReadError(port, "unsupported #-syntax: #%C", c1);
             }
@@ -1076,13 +1082,15 @@ static ScmObj process_sharp_comma(ScmPort *port, ScmObj key, ScmObj args,
     ScmHashEntry *e;
     ScmObj r;
 
+    if (ctx->flags & SCM_READ_DISABLE_CTOR) return SCM_FALSE;
+
     (void)SCM_INTERNAL_MUTEX_LOCK(readCtorData.mutex);
     e = Scm_HashTableGet(readCtorData.table, key);
     (void)SCM_INTERNAL_MUTEX_UNLOCK(readCtorData.mutex);
 
     if (e == NULL) Scm_ReadError(port, "unknown #,-key: %S", key);
     SCM_ASSERT(SCM_PAIRP(e->value));
-    r = Scm_Apply(SCM_CAR(e->value), args);
+    r = Scm_ApplyRec(SCM_CAR(e->value), args);
     if (has_ref) ref_push(ctx, r, SCM_CDR(e->value));
     return r;
 }
@@ -1107,13 +1115,14 @@ static ScmObj maybe_uvector(ScmPort *port, char ch, ScmReadContext *ctx)
 
     c1 = Scm_GetcUnsafe(port);
     if (ch == 'f') {
-        if (c1 != '3' && c1 != '6') {
+        if (c1 != '1' && c1 != '3' && c1 != '6') {
             Scm_UngetcUnsafe(c1, port);
             return SCM_FALSE;
         }
         c2 = Scm_GetcUnsafe(port);
         if (c1 == '3' && c2 == '2') tag = "f32";
         else if (c1 == '6' && c2 == '4') tag = "f64";
+        else if (c1 == '1' && c2 == '6') tag = "f16";
     } else {
         if (c1 == '8') tag = (ch == 's')? "s8" : "u8";
         else if (c1 == '1') {
@@ -1145,7 +1154,8 @@ static ScmObj maybe_uvector(ScmPort *port, char ch, ScmReadContext *ctx)
         /* Require srfi-4 (gauche/uvector)
            NB: we don't need mutex here, for the loading of srfi-4 is
            serialized in Scm_Require. */
-        Scm_Require(SCM_MAKE_STR("gauche/uvector"));
+        Scm_Require(SCM_MAKE_STR("gauche/uvector"),
+                    SCM_LOAD_PROPAGATE_ERROR, NULL);
         if (Scm_ReadUvectorHook == NULL)
             Scm_ReadError(port, "couldn't load srfi-4 module");
     }
