@@ -1,7 +1,7 @@
 /*
  * signal.c - signal handling
  *
- *   Copyright (c) 2000-2006 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2007  Shiro Kawai  <shiro@acm.org>
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: signal.c,v 1.41 2006/11/04 09:56:59 shirok Exp $
+ *  $Id: signal.c,v 1.51 2007/03/02 07:39:14 shirok Exp $
  */
 
 #include <stdlib.h>
@@ -78,7 +78,12 @@
 
 /* Master signal handler vector. */
 static struct sigHandlersRec {
-    ScmObj handlers[NSIG];      /* Scheme signal handlers */
+    ScmObj handlers[NSIG];      /* Scheme signal handlers.  This is #f on
+                                   signals to which Gauche does not install
+                                   C-level signal handler (sig_handle). */
+    ScmSysSigset *masks[NSIG];  /* Signal masks during executing Scheme
+                                   handlers.  Can be NULL, which means
+                                   the handling signal(s) are blocked. */
     sigset_t masterSigset;      /* The signals Gauche is _allowed_ to handle.
                                    set by Scm_SetMasterSigset.
                                    For some signals in this set Gauche sets
@@ -89,6 +94,18 @@ static struct sigHandlersRec {
     ScmInternalMutex mutex;
 } sigHandlers = {{NULL}};
 
+/* Maximum # of the same signals before it is processed by the VM loop.
+   If any one of signals exceeds this count, Gauche exits with Scm_Abort.
+   It is useful to terminate unresponsive program that are executing
+   long-running C-routine and do not returns to VM.
+   The actual limit can be changed at runtime by Scm_SetSignalPendingLimit().
+   If signalPendingLimit is 0, the number of pending signals is unlimited. */
+#define SIGNAL_PENDING_LIMIT_DEFALT 3
+#define SIGNAL_PENDING_LIMIT_MAX 255
+
+static unsigned int signalPendingLimit = SIGNAL_PENDING_LIMIT_DEFALT;
+
+
 /* Table of signals and its initial behavior. */
 #define SIGDEF_NOHANDLE 0       /* Gauche doesn't install a signal handler,
                                    leaving it to the application. */
@@ -193,6 +210,10 @@ static struct sigdesc {
     { NULL, -1 }
 };
 
+/*===============================================================
+ * Signal set operations
+ */
+
 /*
  * utilities for sigset
  */
@@ -242,7 +263,212 @@ ScmObj Scm_SignalName(int signum)
 }
 
 /*
- * default handler
+ * sigset class
+ */
+
+static void sigset_print(ScmObj obj, ScmPort *out, ScmWriteContext *ctx);
+static ScmObj sigset_allocate(ScmClass *klass, ScmObj initargs);
+
+SCM_DEFINE_BUILTIN_CLASS(Scm_SysSigsetClass, sigset_print,
+                         NULL, NULL, sigset_allocate, SCM_CLASS_DEFAULT_CPL);
+
+void sigset_print(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
+{
+    Scm_Printf(out, "#<sys-sigset [");
+    display_sigset(&SCM_SYS_SIGSET(obj)->set, out);
+    Scm_Printf(out, "]>");
+}
+
+ScmObj sigset_allocate(ScmClass *klass, ScmObj initargs)
+{
+    ScmSysSigset *s = SCM_ALLOCATE(ScmSysSigset, klass);
+    SCM_SET_CLASS(s, klass);
+    sigemptyset(&s->set);
+    return SCM_OBJ(s);
+}
+
+ScmSysSigset *make_sigset(void)
+{
+    return SCM_SYS_SIGSET(sigset_allocate(SCM_CLASS_SYS_SIGSET, SCM_NIL));
+}
+
+/* multifunction on sigset
+    if delp == FALSE, signals are added to set.
+    else, signals are removed from set.
+    signals is a list of either integer or #t (all signals), or other sigset.
+*/
+ScmObj Scm_SysSigsetOp(ScmSysSigset *set, ScmObj signals, int delp)
+{
+    ScmObj cp;
+    
+    if (!SCM_PAIRP(signals)) {
+        Scm_Error("list of signals required, but got %S", signals);
+    }
+    SCM_FOR_EACH(cp, signals) {
+        ScmObj s = SCM_CAR(cp);
+        if (SCM_TRUEP(s)) {
+            if (!delp) sigfillset(&set->set);
+            else       sigemptyset(&set->set);
+            break;
+        }
+        if (SCM_SYS_SIGSET_P(s)) {
+            sigset_op(&set->set, &SCM_SYS_SIGSET(s)->set, delp);
+            continue;
+        }
+        if (!SCM_INTP(s) || !validsigp(SCM_INT_VALUE(s))) {
+            Scm_Error("bad signal number %S", s);
+        }
+        if (!delp) sigaddset(&set->set, SCM_INT_VALUE(s));
+        else       sigdelset(&set->set, SCM_INT_VALUE(s));
+    }
+    return SCM_OBJ(set);
+}
+
+/* fill or empty sigset. */
+ScmObj Scm_SysSigsetFill(ScmSysSigset *set, int emptyp)
+{
+    if (emptyp) sigemptyset(&(set->set));
+    else        sigfillset(&(set->set));
+    return SCM_OBJ(set);
+}
+
+/*=============================================================
+ * C-level signal handling
+ */
+
+
+/*-------------------------------------------------------------------
+ * C-level signal handler - just records the signal delivery.
+ */
+
+static void sig_handle(int signum)
+{
+    ScmVM *vm = Scm_VM();
+    /* It is possible that vm == NULL at this point, if the thread is
+       terminating and in the cleanup phase. */
+    if (vm == NULL) return;
+
+    if (signalPendingLimit == 0) {
+        vm->sigq.sigcounts[signum] = 1;
+    } else if (++vm->sigq.sigcounts[signum] >= signalPendingLimit) {
+        Scm_Abort("Received too many signals before processing them.  Exitting for the emergency...\n");
+    }
+    vm->queueNotEmpty |= SCM_VM_SIGQ_MASK;
+}
+
+/*-------------------------------------------------------------------
+ * Signal queue operations
+ */
+
+/*
+ * Clear the signal queue
+ */
+void Scm_SignalQueueClear(ScmSignalQueue* q)
+{
+    int i;
+    for (i=0; i<NSIG; i++) q->sigcounts[i] = 0;
+}
+
+/*
+ * Initializes signal queue
+ */
+void Scm_SignalQueueInit(ScmSignalQueue* q)
+{
+    Scm_SignalQueueClear(q);
+    q->pending = SCM_NIL;
+}
+
+/*
+ * Get/Set signal pending limit
+ */
+int Scm_GetSignalPendingLimit(void)
+{
+    return signalPendingLimit;
+}
+
+void Scm_SetSignalPendingLimit(int num)
+{
+    if (num < 0 || num >= SIGNAL_PENDING_LIMIT_MAX) {
+        Scm_Error("signal-pending-limit argument out of range: %d", num);
+    }
+    signalPendingLimit = num;
+}
+
+/*
+ * Called from VM's safe point to flush the queued signals.
+ * VM already checks there's a pending signal in the queue.
+ */
+void Scm_SigCheck(ScmVM *vm)
+{
+    ScmObj tail, cell, sp;
+    ScmSignalQueue *q = &vm->sigq;
+    sigset_t omask;
+    int i;
+    unsigned char sigcounts[NSIG]; /* copy of signal counter */
+
+    /* Copy VM's signal counter to local storage, for we can't call
+       storage allocation during blocking signals. */
+    SIGPROCMASK(SIG_BLOCK, &sigHandlers.masterSigset, &omask);
+    memcpy(sigcounts, vm->sigq.sigcounts, NSIG * sizeof(unsigned char));
+    Scm_SignalQueueClear(&vm->sigq);
+    vm->queueNotEmpty &= ~SCM_VM_SIGQ_MASK;
+    SIGPROCMASK(SIG_SETMASK, &omask, NULL);
+
+    /* Now, prepare queued signal handlers
+       If an error is thrown in this loop, the queued signals will be
+       lost---it doesn't look like so, but I may overlook something. */
+    tail = q->pending;
+    if (!SCM_NULLP(tail)) tail = Scm_LastPair(tail);
+    for (i=0; i<NSIG; i++) {
+        if (sigcounts[i] == 0) continue;
+        if (SCM_PROCEDUREP(sigHandlers.handlers[i])) {
+            cell = Scm_Cons(SCM_LIST3(sigHandlers.handlers[i],
+                                      SCM_MAKE_INT(i),
+                                      SCM_OBJ_SAFE(sigHandlers.masks[i])),
+                            SCM_NIL);
+            if (SCM_NULLP(tail)) {
+                q->pending = tail = cell;
+            } else {
+                SCM_SET_CDR(tail, cell);
+                tail = SCM_CDR(tail);
+            }
+        }
+    }
+    
+    /* Call the queued signal handlers.  If an error is thrown in one
+       of those handlers, the rest of handlers remain in the queue. */
+    /* TODO: if VM is active, it'd be better to make the active VM to handle
+       those handler procs, instead of calling Scm_Eval. */
+    SCM_FOR_EACH(sp, q->pending) {
+        ScmObj e = SCM_CAR(sp), handler, num, mask;
+        q->pending = SCM_CDR(sp);
+        handler = SCM_CAR(e);
+        num = SCM_CADR(e);
+        mask = SCM_CAR(SCM_CDDR(e));
+        if (SCM_SYS_SIGSET_P(mask)) {
+            sigset_t omask;
+            SCM_UNWIND_PROTECT {
+                SIGPROCMASK(SIG_BLOCK, &SCM_SYS_SIGSET(mask)->set, &omask);
+                Scm_ApplyRec(handler, SCM_LIST1(num));
+            }
+            SCM_WHEN_ERROR {
+                SIGPROCMASK(SIG_SETMASK, &omask, NULL);
+                SCM_NEXT_HANDLER;
+            }
+            SCM_END_PROTECT;
+            SIGPROCMASK(SIG_SETMASK, &omask, NULL);
+        } else {
+            Scm_ApplyRec(handler, SCM_LIST1(num));
+        }
+    }
+}
+
+/*=============================================================
+ * Scheme-level signal handling
+ */
+
+/*-------------------------------------------------------------
+ * Default Scheme-level handlers
  */
 /* For most signals, default handler raises an error. */
 static ScmObj default_sighandler(ScmObj *args, int nargs, void *data)
@@ -317,180 +543,27 @@ static SCM_DEFINE_SUBR(through_sighandler_stub, 1, 0,
 #endif
 
 /*
- * sigset class
+ * An emulation stub for Windows/MinGW
  */
-
-static void sigset_print(ScmObj obj, ScmPort *out, ScmWriteContext *ctx);
-static ScmObj sigset_allocate(ScmClass *klass, ScmObj initargs);
-
-SCM_DEFINE_BUILTIN_CLASS(Scm_SysSigsetClass, sigset_print,
-                         NULL, NULL, sigset_allocate, SCM_CLASS_DEFAULT_CPL);
-
-void sigset_print(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
+#ifdef __MINGW32__
+int sigaction(int signum, const struct sigaction *act,
+	      struct sigaction *oact)
 {
-    Scm_Printf(out, "#<sys-sigset [");
-    display_sigset(&SCM_SYS_SIGSET(obj)->set, out);
-    Scm_Printf(out, "]>");
-}
-
-ScmObj sigset_allocate(ScmClass *klass, ScmObj initargs)
-{
-    ScmSysSigset *s = SCM_ALLOCATE(ScmSysSigset, klass);
-    SCM_SET_CLASS(s, klass);
-    sigemptyset(&s->set);
-    return SCM_OBJ(s);
-}
-
-ScmSysSigset *make_sigset(void)
-{
-    return SCM_SYS_SIGSET(sigset_allocate(SCM_CLASS_SYS_SIGSET, SCM_NIL));
-}
-
-/* multifunction on sigset
-    if delp == FALSE, signals are added to set.
-    else, signals are removed from set.
-    signals is a list of either integer or #t (all signals), or other sigset.
-*/
-ScmObj Scm_SysSigsetOp(ScmSysSigset *set, ScmObj signals, int delp)
-{
-    ScmObj cp;
-    
-    if (!SCM_PAIRP(signals)) {
-        Scm_Error("list of signals required, but got %S", signals);
+    if (oact != NULL) {
+	Scm_Panic("sigaction() with oldact != NULL isn't supported on MinGW port");
     }
-    SCM_FOR_EACH(cp, signals) {
-        ScmObj s = SCM_CAR(cp);
-        if (SCM_TRUEP(s)) {
-            if (!delp) sigfillset(&set->set);
-            else       sigemptyset(&set->set);
-            break;
-        }
-        if (SCM_SYS_SIGSET_P(s)) {
-            sigset_op(&set->set, &SCM_SYS_SIGSET(s)->set, delp);
-            continue;
-        }
-        if (!SCM_INTP(s) || !validsigp(SCM_INT_VALUE(s))) {
-            Scm_Error("bad signal number %S", s);
-        }
-        if (!delp) sigaddset(&set->set, SCM_INT_VALUE(s));
-        else       sigdelset(&set->set, SCM_INT_VALUE(s));
-    }
-    return SCM_OBJ(set);
-}
-
-/* fill or empty sigset. */
-ScmObj Scm_SysSigsetFill(ScmSysSigset *set, int emptyp)
-{
-    if (emptyp) sigemptyset(&(set->set));
-    else        sigfillset(&(set->set));
-    return SCM_OBJ(set);
-}
-
-/*
- * Signal handling
- */
-
-/*
- * System's signal handler - just enqueue the signal
- */
-
-static void sig_handle(int signum)
-{
-    ScmVM *vm = Scm_VM();
-    /* It is possible that vm == NULL at this point, if the thread is
-       terminating and in the cleanup phase. */
-    if (vm == NULL) return;
-    
-    if (vm->sigq.overflow) {
-        /* TODO: we should have better fallback strategy. */
-        struct sigdesc *desc = sigDesc;
-        const char *signame = "unknown signal";
-        for (; desc->name; desc++) {
-            if (desc->num == signum) {
-                signame = desc->name;
-                break;
-            }
-        }
-        Scm_Warn("Signal queue overflow by %s(%d)\n", signame, signum);
-        return;
-    }
-    vm->sigq.queue[vm->sigq.tail++] = signum;
-    if (vm->sigq.tail >= SCM_VM_SIGQ_SIZE) {
-        vm->sigq.tail = 0;
-    }
-    if (vm->sigq.tail == vm->sigq.head) vm->sigq.overflow++;
-    vm->queueNotEmpty |= SCM_VM_SIGQ_MASK;
-}
-
-/*
- * Initializes signal queue
- */
-void Scm_SignalQueueInit(ScmSignalQueue* q)
-{
-    int i;
-    for (i=0; i<SCM_VM_SIGQ_SIZE; i++) q->queue[i] = -1;
-    q->head = q->tail = 0;
-    q->overflow = 0;
-    q->pending = SCM_NIL;
-}
-
-/*
- * Called from VM's safe point to flush the queued signals.
- * VM already checks there's a pending signal in the queue.
- */
-void Scm_SigCheck(ScmVM *vm)
-{
-    ScmObj tail, cell, sp;
-    ScmSignalQueue *q = &vm->sigq;
-    sigset_t omask;
-    int sigQsize, i;
-    int sigQcopy[SCM_VM_SIGQ_SIZE]; /* copy of signal queue */
-
-    /* Copy VM's signal queue to local storage, for we can't call
-       storage allocation during blocking signals. */
-    SIGPROCMASK(SIG_BLOCK, &sigHandlers.masterSigset, &omask);
-    for (sigQsize = 0; q->head != q->tail; sigQsize++) {
-        sigQcopy[sigQsize] = q->queue[q->head++];
-        if (q->head >= SCM_VM_SIGQ_SIZE) q->head = 0;
-    }
-    q->overflow = 0; /*TODO: we should do something*/
-    vm->queueNotEmpty &= ~SCM_VM_SIGQ_MASK;
-    SIGPROCMASK(SIG_SETMASK, &omask, NULL);
-
-    /* Now, prepare queued signal handlers
-       If an error is thrown in this loop, the queued signals will be
-       lost---it doesn't look like so, but I may overlook something. */
-    tail = q->pending;
-    if (!SCM_NULLP(tail)) tail = Scm_LastPair(tail);
-    for (i=0; i<sigQsize; i++) {
-        if (SCM_PROCEDUREP(sigHandlers.handlers[sigQcopy[i]])) {
-            cell = Scm_Acons(sigHandlers.handlers[sigQcopy[i]],
-                             SCM_MAKE_INT(sigQcopy[i]),
-                             SCM_NIL);
-            if (SCM_NULLP(tail)) {
-                q->pending = tail = cell;
-            } else {
-                SCM_SET_CDR(tail, cell);
-                tail = SCM_CDR(tail);
-            }
-        }
-    }
-    
-    /* Call the queued signal handlers.  If an error is thrown in one
-       of those handlers, the rest of handlers remain in the queue. */
-    /* TODO: if VM is active, it'd be better to make the active VM to handle
-       those handler procs, instead of calling Scm_Eval. */
-    SCM_FOR_EACH(sp, q->pending) {
-        ScmObj h = SCM_CAR(sp);
-        q->pending = SCM_CDR(sp);
-        Scm_ApplyRec(SCM_CAR(h), SCM_LIST1(SCM_CDR(h)));
+    if (signal(signum, act->sa_handler) == SIG_ERR) {
+	return -1;
+    } else {
+	return 0;
     }
 }
+#endif /* __MINGW32__ */
 
 /*
  * set-signal-handler!
  */
-ScmObj Scm_SetSignalHandler(ScmObj sigs, ScmObj handler)
+ScmObj Scm_SetSignalHandler(ScmObj sigs, ScmObj handler, ScmSysSigset *mask)
 {
     struct sigaction act;
     struct sigdesc *desc;
@@ -509,6 +582,12 @@ ScmObj Scm_SetSignalHandler(ScmObj sigs, ScmObj handler)
     } else {
         Scm_Error("bad signal number: must be an integer signal number or a <sys-sigset> object, but got %S", sigs);
     }
+
+    if (mask == NULL) {
+        /* If no mask is specified, block singals in SIGS. */
+        mask = make_sigset();
+        mask->set = sigset;
+    }
     
     (void)SCM_INTERNAL_MUTEX_LOCK(sigHandlers.mutex);
     if (SCM_TRUEP(handler)) {
@@ -522,7 +601,7 @@ ScmObj Scm_SetSignalHandler(ScmObj sigs, ScmObj handler)
         badproc = TRUE;
     }
     if (!badproc) {
-        sigemptyset(&act.sa_mask);
+        sigfillset(&act.sa_mask); /* we should block all the signals */
         act.sa_flags = 0;
         for (desc=sigDesc; desc->name; desc++) {
             if (!sigismember(&sigset, desc->num)) continue;
@@ -531,6 +610,7 @@ ScmObj Scm_SetSignalHandler(ScmObj sigs, ScmObj handler)
                 sigactionfailed = desc->num;
             } else {
                 sigHandlers.handlers[desc->num] = handler;
+                sigHandlers.masks[desc->num] = mask;
             }
         }
     }
@@ -542,14 +622,22 @@ ScmObj Scm_SetSignalHandler(ScmObj sigs, ScmObj handler)
 
 ScmObj Scm_GetSignalHandler(int signum)
 {
-    ScmObj r;
     if (signum < 0 || signum >= NSIG) {
         Scm_Error("bad signal number: %d", signum);
     }
-    (void)SCM_INTERNAL_MUTEX_LOCK(sigHandlers.mutex);
-    r = sigHandlers.handlers[signum];
-    (void)SCM_INTERNAL_MUTEX_UNLOCK(sigHandlers.mutex);
-    return r;
+    /* No lock; atomic pointer access */
+    return sigHandlers.handlers[signum];
+}
+
+ScmObj Scm_GetSignalHandlerMask(int signum)
+{
+    ScmSysSigset *r;
+    if (signum < 0 || signum >= NSIG) {
+        Scm_Error("bad signal number: %d", signum);
+    }
+    /* No lock; atomic pointer access */
+    r = sigHandlers.masks[signum];
+    return r? SCM_OBJ(r) : SCM_FALSE;
 }
 
 ScmObj Scm_GetSignalHandlers(void)
@@ -642,6 +730,10 @@ void Scm_SetMasterSigmask(sigset_t *set)
     Scm_VM()->sigMask = sigHandlers.masterSigset;
 }
 
+/*============================================================
+ * Other signal-related operations
+ */
+
 /*
  * set signal mask
  */
@@ -695,7 +787,7 @@ static void scm_sigsuspend(sigset_t *mask)
     ScmVM *vm = Scm_VM();
     for (;;) {
         SIGPROCMASK(SIG_BLOCK, &sigHandlers.masterSigset, &omask);
-        if (vm->sigq.tail != vm->sigq.head) {
+        if (vm->queueNotEmpty & SCM_VM_SIGQ_MASK) {
             SIGPROCMASK(SIG_SETMASK, &omask, NULL);
             Scm_SigCheck(vm);
             continue;
@@ -732,27 +824,96 @@ ScmObj Scm_Pause(void)
 }
 
 /*
- * Emulation stubs for Windows/MinGW
+ * Sigwait wrapper
+ *
+ * The behavior of sigwait is undefined if a signal handler is set to
+ * the waiting signal.  On Cygwin, for example, using both signal handler
+ * and sigwait makes havoc.  Since Gauche installs sig_handle()
+ * implicitly to some signals, a casual user may be confused by the
+ * unpredictable behavior when he doesn't reset signal handlers explicitly.
+ * So we take care of them here. 
+ *
+ * We remove the signal handlers for the signals to be waited before calling
+ * sigwait(), and restore them after its return.  We assume those signals
+ * are blocked at this moment (if not, the behavior of sigwait() is
+ * undefined), so we don't need to care about race condition.  If another
+ * thread replaces signal handlers during this thread's waiting for a
+ * signal, it would be reverted upon returning from this function, but 
+ * such operation is inherently unsafe anyway, so we don't care.
  */
-#ifdef __MINGW32__
-
-int sigaction(int signum, const struct sigaction *act,
-	      struct sigaction *oact)
+int Scm_SigWait(ScmSysSigset *mask)
 {
-    if (oact != NULL) {
-	Scm_Panic("sigaction() with oldact != NULL isn't supported on MinGW port");
+#if defined(HAVE_SIGWAIT)
+    int i, r = 0, sig = 0;
+    int failed_sig = -1;
+    int sigwait_called = FALSE;
+    int errno_save = 0;
+    sigset_t to_wait;        /* real set of signals to wait */
+    sigset_t saved;
+    struct sigaction act, oacts[NSIG];
+
+    (void)SCM_INTERNAL_MUTEX_LOCK(sigHandlers.mutex);
+    /* we can't wait for the signals Gauche doesn't handle. */
+    to_wait = mask->set;
+    for (i=0; i<NSIG; i++) {
+        if (!sigismember(&sigHandlers.masterSigset, i)) {
+            sigdelset(&to_wait, i);
+        }
     }
-    if (signal(signum, act->sa_handler) == SIG_ERR) {
-	return -1;
-    } else {
-	return 0;
+
+    /* Remove C-level handlers */
+    sigemptyset(&saved);
+    act.sa_handler = SIG_DFL;
+    act.sa_flags = 0;
+    for (i=1; i<NSIG; i++) {
+        if (!sigismember(&to_wait, i)) continue;
+        if (sigaction(i, &act, &oacts[i]) < 0) {
+            failed_sig = i;
+            errno_save = errno;
+            break;
+        }
+        sigaddset(&saved, i);
     }
+    
+    if (failed_sig < 0) {
+        (void)SCM_INTERNAL_MUTEX_UNLOCK(sigHandlers.mutex);
+        sigwait_called = TRUE;
+        r = sigwait(&to_wait, &sig);
+        (void)SCM_INTERNAL_MUTEX_LOCK(sigHandlers.mutex);
+    }
+
+    /* Restore C-level handlers */
+    for (i=1; i<NSIG; i++) {
+        if (!sigismember(&saved, i)) continue;
+        if (sigaction(i, &oacts[i], NULL) < 0) {
+            failed_sig = i;
+            errno_save = errno;
+        }
+    }
+    (void)SCM_INTERNAL_MUTEX_UNLOCK(sigHandlers.mutex);
+
+    /* error handling */
+    if (failed_sig >= 0) {
+        errno = errno_save;
+        Scm_SysError("sigaction(2) call failed on signal %d"
+                     " %s sigwait call",
+                     failed_sig,
+                     sigwait_called? "after" : "before");
+    }
+    if (r != 0) {
+        errno = r;
+        Scm_SysError("sigwait failed");
+    }
+    return sig;
+#else  /* !HAVE_SIGWAIT */
+    Scm_Error("sigwait not supported on this platform");
+    return 0;
+#endif
 }
-#endif /* __MINGW32__ */
 
 
-/*
- * initialize
+/*================================================================
+ * Initialize
  */
 
 void Scm__InitSignal(void)
