@@ -1,7 +1,7 @@
 /*
  * system.c - system interface
  *
- *   Copyright (c) 2000-2006 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2007  Shiro Kawai  <shiro@acm.org>
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: system.c,v 1.84 2006/11/11 07:48:52 shirok Exp $
+ *  $Id: system.c,v 1.88 2007/04/19 05:52:05 shirok Exp $
  */
 
 #include <stdio.h>
@@ -196,7 +196,7 @@ ScmObj Scm_GlobDirectory(ScmString *pattern)
     glob_t globbed;
     ScmObj head = SCM_NIL, tail = SCM_NIL;
     int i, r;
-    SCM_SYSCALL(r, glob(Scm_GetStringConst(pattern), 0, NULL, &globbed));
+    r = glob(Scm_GetStringConst(pattern), 0, NULL, &globbed);
     if (r) {
         globfree(&globbed);
 #if defined(GLOB_NOMATCH)
@@ -392,11 +392,11 @@ static void copy_win32_path(ScmDString *dst,
     while (srcp < end) {
         ScmChar ch;
         if (*srcp == '/' || *srcp == '\\') {
-            Scm_DStringPutc(dst, SEPARATOR);
+            ch = SEPARATOR;
         } else {
             SCM_CHAR_GET(srcp, ch);
-            Scm_DStringPutc(dst, ch);
         }
+        Scm_DStringPutc(dst, ch);
         srcp += SCM_CHAR_NBYTES(ch);
     }
 }
