@@ -1,6 +1,6 @@
-;; this test only works when the core system is compiled with euc-jp.
+;; -*- coding: euc-jp -*-
 
-;; $Id: euc-jp.scm,v 1.21 2003/03/25 06:18:39 shirok Exp $
+;; $Id: euc-jp.scm,v 1.23 2007/03/02 07:20:42 shirok Exp $
 
 (use gauche.test)
 
@@ -71,15 +71,6 @@
 (test* "string-scan" #f
        (string-scan "" ""))
 
-(test* "string-substitute!" "defghi"
-       (let ((s (string-copy "abcdefghi")))
-         (string-substitute! s 0 "")
-         s))
-(test* "string-substitute!" "abcghi"
-       (let ((s (string-copy "abcdefghi")))
-         (string-substitute! s 3 "")
-         s))
-
 ;;-------------------------------------------------------------------
 (test-section "string-pointer")
 (define sp #f)
@@ -147,9 +138,6 @@
 (test* "string-append" 10
        (string-length (string-append "" "" #*"")))
 
-(test* "string-substitute!" #*"\xa4bc\xa4"
-       (string-substitute! (string-copy #*"") 1 #*"bc"))
-
 (test* "string-incompltet->incomplete" ""
        (string-incomplete->complete
         (string-append #*"\xa4" #*"\xa2")))
