@@ -239,6 +239,92 @@
        (with-input-from-file "tmp1.o"
          (lambda () (port-map (lambda (x) x) read))))
 
+;;-------------------------------------------------------------------
+(test-section "with-ports")
+
+(test* "with-input-from-port" '(#\b #\d #\c #\a)
+       (let ((x (open-input-string "ab"))
+             (y (open-input-string "cd"))
+             (r '())
+             (restart #f))
+         (if (call/cc 
+              (lambda (escape)
+                (with-input-from-port x
+                  (lambda ()
+                    (push! r (read-char))
+                    (with-input-from-port y
+                      (lambda ()
+                        (push! r (read-char))
+                        (call/cc (lambda (k) (set! restart k) (escape #t)))
+                        (push! r (read-char))))
+                    (push! r (read-char))))
+                #f))
+           (restart #f)
+           r)))
+
+(for-each
+ (lambda (with cur name)
+   (test* "with-,|name|-to-port" '("ad" "bc")
+          (let ((x (open-output-string))
+                (y (open-output-string))
+                (restart #f))
+            (if (call/cc 
+                 (lambda (escape)
+                   (with x
+                     (lambda ()
+                       (write-char #\a (cur))
+                       (with y
+                         (lambda ()
+                           (write-char #\b (cur))
+                           (call/cc (lambda (k) (set! restart k) (escape #t)))
+                           (write-char #\c (cur))))
+                       (write-char #\d (cur))))
+                   #f))
+              (restart #f)
+              (list (get-output-string x) (get-output-string y))))))
+ `(,with-output-to-port ,with-error-to-port)
+ `(,current-output-port ,current-error-port)
+ '("output" "error"))
+
+(test* "with-ports 1" '("a" "b")
+       (let ((o0 (open-output-string))
+             (o1 (open-output-string)))
+         (with-ports (open-input-string "abcd") o0 o1
+           (lambda ()
+             (write-char (read-char))
+             (write-char (read-char) (current-error-port))))
+         (list (get-output-string o0) (get-output-string o1))))
+(test* "with-ports 2" '("B" "A")
+       (let ((o0 (open-output-string))
+             (o1 (open-output-string)))
+         (with-ports (open-input-string "abcd") o0 o0
+           (lambda ()
+             (with-ports (open-input-string "ABCD") o1 #f
+               (lambda ()
+                 (write-char (read-char))
+                 (write-char (read-char) (current-error-port))))))
+         (list (get-output-string o0) (get-output-string o1))))
+(test* "with-ports 3" '("A" "B")
+       (let ((o0 (open-output-string))
+             (o1 (open-output-string)))
+         (with-ports (open-input-string "abcd") o0 o0
+           (lambda ()
+             (with-ports (open-input-string "ABCD") #f o1
+               (lambda ()
+                 (write-char (read-char))
+                 (write-char (read-char) (current-error-port))))))
+         (list (get-output-string o0) (get-output-string o1))))
+(test* "with-ports 4" '("" "ab")
+       (let ((o0 (open-output-string))
+             (o1 (open-output-string)))
+         (with-ports (open-input-string "abcd") o0 o0
+           (lambda ()
+             (with-ports #f o1 o1
+               (lambda ()
+                 (write-char (read-char))
+                 (write-char (read-char) (current-error-port))))))
+         (list (get-output-string o0) (get-output-string o1))))
+
 ;;-------------------------------------------------------------------
 (test-section "seeking")
 
@@ -580,5 +666,91 @@
 (test* "#;#;(123 456) 789 1" 1
        (read-from-string "#;#;(123 456) 789 1"))
 
+(define *counter* 0)
+
+(define-reader-ctor 'countup
+  (lambda () (inc! *counter*) #t))
+
+(test* "S-expression comment and read-time constructor" 2
+       (begin (list #,(countup) #;#,(countup) #,(countup))
+              *counter*))
+
+;;-------------------------------------------------------------------
+(test-section "coding-aware-port basic")
+
+;; Testing source port _without_ any conversion.  Basically, these
+;; tests just checks up the boundary condition of source-port prefetching
+;; routine.
+;; The actual conversion is tested in ext/charconv.
+
+(define (with-coding-aware-port input proc)
+  (let* ((src  (open-input-string input))
+         (wrap (open-coding-aware-port src)))
+    (proc src wrap)))
+
+(test* "ownership" #t
+       (with-coding-aware-port
+        "abc"
+        (lambda (src wrap)       
+          (close-input-port wrap)          
+          (port-closed? src))))
+
+(test* "read from empty port" '(#t #t #t #t #t)
+       (map (lambda (p) (with-coding-aware-port "" p))
+            (list
+             (lambda (src wrap) (eof-object? (read-char wrap)))
+             (lambda (src wrap) (eof-object? (read-byte wrap)))
+             (lambda (src wrap) (eof-object? (peek-char wrap)))
+             (lambda (src wrap) (eof-object? (peek-byte wrap)))
+             (lambda (src wrap) (eof-object? (read-line wrap))))))
+
+(let ((tdata '("abc" "abc\n" "\nabc" "abc\ndef\n"
+               "abc\ndef\nghi" "abc\ndef\nghi\n"
+               "abc\r\ndef\r\nghi" "abc\r\ndef\r\nghi\r\n"
+               "abc\rdef\rghi" "abc\rdef\rghi\r"
+               "abc\ndef\nghi\njkl" "abc\ndef\nghi\njkl\n")))
+  (test* "read from simple contents"
+         tdata
+         (map (lambda (i)
+                (with-coding-aware-port
+                 i
+                 (lambda (src wrap)
+                   (let loop ((ch (read-char wrap))
+                              (r  '()))
+                     (if (eof-object? ch)
+                       (list->string (reverse r))
+                       (loop (read-char wrap) (cons ch r)))))))
+              tdata)))
+
+(let ((tdata '("coding: abcde\naa"
+               ";coding:\nabcdef\naa"
+               ";coding:\r\nabcdef\r\naa"
+               ";coding:\rabcdef\raa"
+               "coding: coding: coding:; abcde\naa"
+               "coding: coding: coding:; abcde\r\naa"
+               ";; co\nding: foobar\naa"
+               ";; co\r\nding: foobar\r\naa"
+               ";; co\rding: foobar\raa"
+               ";; coding:\n;; foobar\naa"
+               ";; coding : foobar\naa"
+               ";; coding : foobar\r\naa"
+               ";; coding : foobar\raa"
+               "\n\n;; coding: foobar\naa"
+               "\r\n\r\n;; coding: foobar\r\naa"
+               "\r\r;; coding: foobar\raa"
+               "\n;;    codincodincoding:\naa")))
+  (test* "to confuse DFA"
+         tdata
+         (map (lambda (i)
+                (with-coding-aware-port
+                 i
+                 (lambda (src wrap)
+                   (let loop ((ch (read-char wrap))
+                              (r  '()))
+                     (if (eof-object? ch)
+                       (list->string (reverse r))
+                       (loop (read-char wrap) (cons ch r)))))))
+              tdata)))
+
 (test-end)
 
